""" Basic demonstation of integrating normals and performing peak detection

"""
import time
import sys
import argparse
import numpy as np
import open3d as o3d

from fastgac import GaussianAccumulatorS2Beta, MatX3d, convert_normals_to_hilbert, IcoCharts
from fastgac.peak_and_cluster import find_peaks_from_accumulator, find_peaks_from_ico_charts
from fastgac.o3d_util import get_arrow, get_pc_all_peaks, get_arrow_normals, plot_meshes, assign_vertex_colors, plot_meshes, get_colors, create_open_3d_mesh
import matplotlib.pyplot as plt

from tests.python.helpers.setup_helper import cluster_normals, sort_by_distance_from_point
np.random.seed(1)

np.set_printoptions(suppress=True, precision=3)

def integrate_normals_and_visualize(to_integrate_normals, ga):
    to_integrate_normals_mat = MatX3d(to_integrate_normals)
    t0 = time.perf_counter()
    neighbors_idx = np.asarray(ga.integrate(to_integrate_normals_mat))
    t1 = time.perf_counter()
    elapsed_time = (t1 - t0) * 1000
    normalized_counts = np.asarray(ga.get_normalized_bucket_counts())
    color_counts = get_colors(normalized_counts)[:, :3]
    refined_icosahedron_mesh = create_open_3d_mesh(np.asarray(ga.mesh.triangles), np.asarray(ga.mesh.vertices))
    # Colorize normal buckets
    colored_icosahedron = assign_vertex_colors(refined_icosahedron_mesh, color_counts, None)

    return colored_icosahedron

    
def example_normals(normals:np.ndarray):
    LEVEL = 4
    kwargs_base = dict(level=LEVEL)
    kwargs_s2 = dict(**kwargs_base)

    axis_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(0.5).translate([-2.0, 0, 0])
    # Create Gaussian Accumulator
    ga_cpp_s2 = GaussianAccumulatorS2Beta(**kwargs_s2)
    # Integrate the normals and get open3d visualization
    colored_icosahedron  = integrate_normals_and_visualize(normals, ga_cpp_s2)
    o3d.visualization.draw_geometries([colored_icosahedron, axis_frame])
    # Create the IcoChart for unwrapping
    ico_chart_ = IcoCharts(LEVEL)
    normalized_bucket_counts_by_vertex = ga_cpp_s2.get_normalized_bucket_counts_by_vertex(True)
    ico_chart_.fill_image(normalized_bucket_counts_by_vertex)
    average_vertex_normals = np.asarray(ga_cpp_s2.get_average_normals_by_vertex(True))

    # 2D Peak Detection
    find_peaks_kwargs = dict(threshold_abs=20, min_distance=1, exclude_border=False, indices=False)
    cluster_kwargs = dict(t=0.05, criterion='distance')
    average_filter = dict(min_total_weight=0.2)

    # New simplified API for finding peaks
    res = np.array(ga_cpp_s2.find_peaks(threshold_abs=find_peaks_kwargs['threshold_abs'], cluster_distance=cluster_kwargs['t'], min_cluster_weight=average_filter['min_total_weight']))
    print("New Detected Peaks:")
    res = sort_by_distance_from_point(res)
    print(res)

    # Old Way of finding peaks
    _, _, avg_peaks, _ = find_peaks_from_ico_charts(ico_chart_, np.asarray(normalized_bucket_counts_by_vertex), vertices=average_vertex_normals, find_peaks_kwargs=find_peaks_kwargs, cluster_kwargs=cluster_kwargs)
    avg_peaks = sort_by_distance_from_point(avg_peaks)
    print("Detected Peaks:")
    print(avg_peaks)

    full_image = np.asarray(ico_chart_.image)
    plt.imshow(full_image)
    plt.xticks(np.arange(0, full_image.shape[1], step=1))
    plt.yticks(np.arange(0, full_image.shape[0], step=1))
    plt.show()

    # Don't forget to reset the GA
    ga_cpp_s2.clear_count()

def main():
    parser = argparse.ArgumentParser(description='Integrate some Normals')
    parser.add_argument('--path', type=str, help='Specify an optional file')
    args = parser.parse_args()
    if args.path is None:
        clusters, normals = cluster_normals(10, 1000, patch_deg=5)
        combined =np.concatenate(clusters)
    else:
        print("loading data from ", args.path)
        data = np.load(args.path)
        combined = data['clusters']
        normals = data['normals']
    print(sort_by_distance_from_point(normals))
    # sys.exit()
    # normals = np.asarray([
    #     [0.0, 0.0, 0.95],
    #     [0.0, 0.0, 0.98],
    #     [0.95, 0.0, 0],
    #     [0.98, 0.0, 0]
    # ])
    example_normals(combined)

if __name__ == "__main__":
    main()


import numpy as np
import math
from scipy import stats
from fastgac import GaussianAccumulatorOpt, GaussianAccumulatorS2, MatX3d
import matplotlib.pyplot as plt
from matplotlib.transforms import Bbox
import seaborn as sns
# import tikzplotlib


# sns.set()
sns.set(font_scale=1.1, style='white')  # crazy big

def set_attributes(axes, xlabel='Array Index', ylabel='Unique ID (Hilbert Value)'):
    for ax in axes:
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.legend(loc='upper left')

def full_extent(ax, pad=0.0):
    """Get the full extent of an axes, including axes labels, tick labels, and
    titles."""
    # For text objects, we need to draw the figure first, otherwise the extents
    # are undefined.
    ax.figure.canvas.draw()
    items = ax.get_xticklabels() + ax.get_yticklabels() 
    items += [ax, ax.title, ax.xaxis.label, ax.yaxis.label]
    items += [ax, ax.title]
    items += [ax.get_xaxis().get_label(), ax.get_yaxis().get_label()]
    bbox = Bbox.union([item.get_window_extent() for item in items])

    return bbox.expanded(1.0 + pad, 1.0 + pad)

def plot_indices(ga_s2):


    indices_s2 = np.asarray(ga_s2.get_bucket_sfc_values())
    indices_s2 = indices_s2 - np.min(indices_s2)
    fig0, ax0 = plt.subplots(1, 1, figsize=(4.5, 4))
    fig1, ax1 = plt.subplots(1, 1, figsize=(4.5, 4))
    fig2, ax2 = plt.subplots(1, 1, figsize=(4.5, 4))


    bucket_positions_s2 = np.arange(indices_s2.shape[0])
    ax0.plot(indices_s2, bucket_positions_s2 ,'.', markersize=2, label='Cells')


    
    ax1.plot(indices_s2, bucket_positions_s2 ,'.', markersize=2, label='Cells')
    predicted, r_value, error = calc_linear_regression(bucket_positions_s2, indices_s2, ax1)
    # ax[1].legend(loc='upper left')

    ax2.plot(indices_s2, bucket_positions_s2 ,'.', markersize=2, label='Cells')
    ax2.plot(indices_s2, predicted, 'g', label=r'Fitted Line, $R^2$ = {:.1}'.format(r_value))
    ax3 = ax2.twinx()
    ax3.set_ylabel(r"Error")
    ax3.plot(indices_s2, error, '-r', label="Error")
    # Dummy line for legend
    ax2.plot(np.nan, '-r', label='Error')
        

    set_attributes([ax0, ax1, ax2], ylabel='Cells Array Index', xlabel='S2ID')
    ax2.set_xlim([6.2e18, 7.2e18])
    ax2.set_ylim([2300, 2700])
    fig0.tight_layout()
    fig1.tight_layout()
    fig2.tight_layout()

    fig0.savefig('assets/imgs/ga_linear_interp_0.pdf', bbox_inches='tight')
    fig1.savefig('assets/imgs/ga_linear_interp_1.pdf', bbox_inches='tight')
    fig2.savefig('assets/imgs/ga_linear_interp_2.pdf', bbox_inches='tight')


    # tikzplotlib.save("assets/imgs/ga_linear_interp_0.tex", figure=fig0)

    # extent = full_extent(ax[0]).transformed(fig.dpi_scale_trans.inverted())
    # fig.savefig('assets/imgs/ga_linear_interp_0.pdf', bbox_inches=extent.expanded(1.0, 1.0))

    # extent = full_extent(ax[1]).transformed(fig.dpi_scale_trans.inverted())
    # fig.savefig('assets/imgs/ga_linear_interp_1.pdf', bbox_inches=extent.expanded(1.0, 1.0))

    # items = [ax2.get_xaxis().get_label().get_window_extent(), ax2.get_yaxis().get_label().get_window_extent()]
    # extent = Bbox.union([full_extent(ax[2]), *items]).transformed(fig.dpi_scale_trans.inverted())
    # fig.savefig('assets/imgs/ga_linear_interp_2.pdf', bbox_inches=extent.expanded(1.0, 1.0))


    # plt.show()

def calc_linear_regression(bucket_positions, indices, ax):
    max_hv = np.max(indices)
    min_hv = np.min(indices)
    print("max, min", max_hv, min_hv)
    slope, intercept, r_value, p_value, std_err = stats.linregress(indices, bucket_positions)
    print("Slope and Intercept", slope, intercept)
    predicted = intercept + slope*indices
    ax.plot(indices, predicted, 'g', label=r'Fitted Line, $R^2$ = {:.1}'.format(r_value))
    error = predicted - bucket_positions

    min_val = math.floor(np.min(error))
    max_val = math.ceil(np.max(error))
    print("Min Index Error: {}; Max Index Error: {}".format(min_val, max_val))
    return predicted, r_value, error

def main():
    kwargs = dict(level=4, max_phi=180)
    ga_s2 = GaussianAccumulatorS2(**kwargs)

    plot_indices(ga_s2)


if __name__ == "__main__":
    main()


import time
from pathlib import Path
from collections import namedtuple

import numpy as np
import open3d as o3d
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt

from fastgac import GaussianAccumulatorKD, GaussianAccumulatorOpt, GaussianAccumulatorS2, MatX3d, convert_normals_to_hilbert, IcoCharts
from fastgac.peak_and_cluster import find_peaks_from_accumulator, find_peaks_from_ico_charts
from fastgac.o3d_util import get_arrow, get_pc_all_peaks, get_arrow_normals, plot_meshes, assign_vertex_colors, create_open_3d_mesh, get_colors

from src.Python.slowga import (GaussianAccumulatorKDPy, filter_normals_by_phi)


THIS_DIR = Path(__file__).parent
FIXTURES_DIR = THIS_DIR / "../../fixtures/"
REALSENSE_DIR = (FIXTURES_DIR / "realsense").absolute()
EXAMPLE_MESH_1 = REALSENSE_DIR / "example_mesh.ply"
EXAMPLE_MESH_2 = REALSENSE_DIR / "dense_first_floor_map.ply"
EXAMPLE_MESH_3 = REALSENSE_DIR / "sparse_basement.ply"

ALL_MESHES = [EXAMPLE_MESH_1, EXAMPLE_MESH_2, EXAMPLE_MESH_3]
ALL_MESHES_ROTATIONS = [None, R.from_rotvec(-np.pi / 2 * np.array([1, 0, 0])),
                        R.from_rotvec(-np.pi / 2 * np.array([1, 0, 0]))]



def get_image_peaks(ga_cpp_s2, level=2, **kwargs):
    ico_chart = IcoCharts(level)
    normalized_bucket_counts_by_vertex = ga_cpp_s2.get_normalized_bucket_counts_by_vertex(True)
    ico_chart.fill_image(normalized_bucket_counts_by_vertex)

    find_peaks_kwargs=dict(threshold_abs=25, min_distance=1, exclude_border=False, indices=False)
    cluster_kwargs=dict(t=0.10, criterion='distance')
    average_filter=dict(min_total_weight=0.10)

    peaks, clusters, avg_peaks, avg_weights = find_peaks_from_ico_charts(ico_chart, np.asarray(normalized_bucket_counts_by_vertex), find_peaks_kwargs, cluster_kwargs, average_filter)
    gaussian_normals_sorted = np.asarray(ico_chart.sphere_mesh.vertices)
    pcd_all_peaks = get_pc_all_peaks(peaks, clusters, gaussian_normals_sorted)
    arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)

    return [pcd_all_peaks, *arrow_avg_peaks]

def plot_hilbert_curve(ga: GaussianAccumulatorKDPy, plot=False):
    normals = np.asarray(ga.get_bucket_normals())
    normalized_counts = np.asarray(ga.get_normalized_bucket_counts())
    colors = get_colors(normalized_counts)[:, :3]
    bucket_normals_hv = np.asarray(ga.get_bucket_sfc_values())
    num_buckets = ga.num_buckets
    idx_sort = np.argsort(bucket_normals_hv)
    bucket_normals_hv_sorted = bucket_normals_hv[idx_sort]
    colors = colors[idx_sort, :]
    accumulator_normalized_sorted = normalized_counts[idx_sort]
    gaussian_normals_sorted = normals[idx_sort, :]
    # print(gaussian_normals_sorted)
    # Find Peaks
    peaks, clusters, avg_peaks, avg_weights = find_peaks_from_accumulator(gaussian_normals_sorted, accumulator_normalized_sorted)

    # print(avg_peaks)

    # 2D Plots
    if plot:
        class_name_str = type(ga).__name__
        if class_name_str == 'GaussianAccumulatorS2':
            fig, ax = plt.subplots(1, 1, figsize=(5, 5))

        else:
            proj_ = np.asarray(ga.get_bucket_projection())
            proj = proj_[idx_sort, :]
            fig, axs = plt.subplots(2, 1, figsize=(8, 10))
            ax = axs[0]
            scatter1 = ax.scatter(proj[:, 0], proj[:, 1], c=colors, label='Projected Buckets')
            scatter2 = ax.scatter(proj[peaks, :][:, 0], proj[peaks, :][:, 1],
                                marker='x', c=clusters, label='Clusters', cmap='tab20')
            ax.set_title("Hilbert Curve with Azimuth Equidistant Projection")
            ax.set_xlabel("x*")
            ax.set_ylabel("y*")
            line1 = ax.plot(proj[:, 0], proj[:, 1], c='k', label='Hilbert Curve Connections')[0]
            ax.axis('equal')
            leg = ax.legend(loc='upper left', fancybox=True, shadow=True)

            # we will set up a dict mapping legend line to orig line, and enable
            # picking on the legend line
            lines = [line1, scatter1, scatter2]
            lined = dict()
            for legline, origline in zip(leg.legendHandles, lines):
                legline.set_picker(5)  # 5 pts tolerance
                lined[legline] = origline

            def onpick(event):
                # on the pick event, find the orig line corresponding to the
                # legend proxy line, and toggle the visibility
                legline = event.artist
                origline = lined[legline]
                vis = not origline.get_visible()
                origline.set_visible(vis)
                # Change the alpha on the line in the legend so we can see what lines
                # have been toggled
                if vis:
                    legline.set_alpha(1.0)
                else:
                    legline.set_alpha(0.2)
                fig.canvas.draw()

            ax = axs[1]
            fig.canvas.mpl_connect('pick_event', onpick)

        ax.bar(np.arange(num_buckets), accumulator_normalized_sorted)
        ax.scatter(peaks, accumulator_normalized_sorted[peaks], marker='x', c=clusters, cmap='tab20')

        ax.set_title("Histogram of Normal Counts sorted by Hilbert Values")
        ax.set_xlabel("Hilbert Value (Ascending)")
        ax.set_ylabel("Normal Counts")
        fig.tight_layout()
        plt.show()

    pcd_all_peaks = get_pc_all_peaks(peaks, clusters, gaussian_normals_sorted)
    arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)
    return [pcd_all_peaks, *arrow_avg_peaks]


def visualize_gaussian_integration(ga: GaussianAccumulatorKDPy, mesh: o3d.geometry.TriangleMesh, ds=50, min_samples=10000, max_phi=100, integrate_kwargs=dict()):
    num_buckets = ga.num_buckets
    to_integrate_normals = np.asarray(mesh.triangle_normals)
    # remove normals on bottom half of sphere
    to_integrate_normals, _ = filter_normals_by_phi(to_integrate_normals, max_phi=180)
    # determine optimal sampling
    num_normals = to_integrate_normals.shape[0]
    ds_normals = int(num_normals / ds)
    to_sample = max(min([num_normals, min_samples]), ds_normals)
    ds_step = int(num_normals / to_sample)
    # perform sampling of normals
    to_integrate_normals = to_integrate_normals[0:num_normals:ds_step, :]
    mask = np.asarray(ga.mask)
    query_size = to_integrate_normals.shape[0]

    mask = np.ones((np.asarray(ga.mesh.triangles).shape[0],), dtype=bool)
    mask[num_buckets:] = False
    class_name_str = type(ga).__name__
    # integrate normals
    if class_name_str in ['GaussianAccumulatorKD', 'GaussianAccumulatorOpt', 'GaussianAccumulatorS2']:
        to_integrate_normals = MatX3d(to_integrate_normals)

        # mask = np.ma.make_mask(mask)

    # triangles = np.asarray(ga.mesh.triangles)
    t0 = time.perf_counter()
    neighbors_idx = np.asarray(ga.integrate(to_integrate_normals, **integrate_kwargs))
    t1 = time.perf_counter()
    elapsed_time = (t1 - t0) * 1000
    print("{}; KD tree size: {}; Query Size (K): {}; Execution Time(ms): {:.1f}".format(
        class_name_str, num_buckets, query_size, elapsed_time))
    normalized_counts = np.asarray(ga.get_normalized_bucket_counts())
    color_counts = get_colors(normalized_counts)[:, :3]
    # print(normalized_counts)

    refined_icosahedron_mesh = create_open_3d_mesh(np.asarray(ga.mesh.triangles), np.asarray(ga.mesh.vertices))

    # Colorize normal buckets
    colored_icosahedron = assign_vertex_colors(refined_icosahedron_mesh, color_counts, mask)
    return colored_icosahedron, np.asarray(to_integrate_normals), neighbors_idx


def create_line_set(normals_sorted):
    normals_o3d = o3d.utility.Vector3dVector(normals_sorted)
    line_idx = [[i, i + 1] for i in range(normals_sorted.shape[0] - 1)]
    line_idx_o3d = o3d.utility.Vector2iVector(line_idx)
    return o3d.geometry.LineSet(normals_o3d, line_idx_o3d)


def plot_issues_2(idx, normals, chosen_buckets, ga, mesh):

    normal_idx = idx[0]
    bad_normals = normals[idx, :]
    bad_chosen_triangles = chosen_buckets[idx]

    chosen_triangle = bad_chosen_triangles[0]
    bad_normal = np.expand_dims(bad_normals[0, :], axis=0)
    print(bad_normal)
    # chosen_normal =

    # Get all bucket normals
    bucket_normals_sorted = np.asarray(ga.get_bucket_normals())

    # Get 2D projection of bad normal
    projected_normals, hv = convert_normals_to_hilbert(MatX3d(normals), ga.projected_bbox)
    projected_normals = np.asarray(projected_normals)
    bad_projected_normal = projected_normals[normal_idx, :]

    # Get Projected coordinates of buckets
    proj = np.asarray(ga.get_bucket_projection())
    normalized_counts = np.asarray(ga.get_normalized_bucket_counts())
    colors = get_colors(normalized_counts)[:, :3]

    # Get Projected Coordinates of Buckets and normals!
    all_normals = np.vstack((bucket_normals_sorted, normals))
    all_projected_normals, all_hv = convert_normals_to_hilbert(MatX3d(all_normals), ga.projected_bbox)
    all_projected_normals = np.asarray(all_projected_normals)
    all_hv = np.asarray(all_hv)

    idx_sort = np.argsort(all_hv)
    all_projected_normals = all_projected_normals[idx_sort]
    all_normals_sorted = np.ascontiguousarray(all_normals[idx_sort])

    fig, axs = plt.subplots(1, 1, figsize=(5, 5))
    ax = axs
    # plot buckets
    scatter1 = ax.scatter(proj[:, 0], proj[:, 1], c=colors, label='Projected Buckets')
    scatter2 = ax.scatter(proj[chosen_triangle, 0], proj[chosen_triangle, 1], c='r', label='Chosen Triangle')
    # scatter3 = ax.scatter(proj[133, 0], proj[133, 1], c='g', label='Hilbert Mapped Triangle')
    scatter4 = ax.scatter(bad_projected_normal[0], bad_projected_normal[1],
                          c=[[0.5, 0.5, 0.5]], label='Projected Normal')
    line1 = ax.plot(all_projected_normals[:, 0], all_projected_normals[:, 1],
                    c='k', label='Hilbert Curve Connections')[0]
    leg = ax.legend(loc='upper left', fancybox=True, shadow=True)
    # scatter2 = ax.scatter(proj[133, 0], proj[133, 1], c='k')
    plt.show()

    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(bad_normal)
    pcd.colors = o3d.utility.Vector3dVector([[0.5, 0.5, 0.5]])

    pcd2 = o3d.geometry.PointCloud()
    pcd2.points = o3d.utility.Vector3dVector(bucket_normals_sorted)
    pcd2.colors = o3d.utility.Vector3dVector([[0.0, 1.0, 0]])

    vertex_colors = np.asarray(mesh.vertex_colors)
    triangles = np.asarray(mesh.triangles)
    p_idx = triangles[chosen_triangle, :]
    vertex_colors[p_idx] = [1, 0, 0]
    # p_idx = triangles[133,:]
    # vertex_colors[p_idx] = [0, 1, 0]

    ls = create_line_set(all_normals_sorted * 1.002)
    o3d.visualization.draw_geometries([mesh, pcd, pcd2, ls])


def reorder_single_hv_to_s2(ga_cpp_kdd, ga_cpp_s2):
    s2_normals = ga_cpp_s2.get_bucket_normals()
    _, hv = convert_normals_to_hilbert(s2_normals, ga_cpp_kdd.projected_bbox)
    hv = np.asarray(hv)
    idx_sort = np.argsort(hv)
    hv = hv[idx_sort]
    return idx_sort


def main():
    kwargs_base = dict(level=4, max_phi=180)
    kwargs_kdd = dict(**kwargs_base, max_leaf_size=10)
    kwargs_opt = dict(**kwargs_base)
    kwargs_s2 = dict(**kwargs_base)

    kwargs_opt_integrate = dict(num_nbr=12)
    # Get an Example Mesh
    ga_py_kdd = GaussianAccumulatorKDPy(**kwargs_kdd)
    ga_cpp_kdd = GaussianAccumulatorKD(**kwargs_kdd)
    ga_cpp_opt = GaussianAccumulatorOpt(**kwargs_opt)
    ga_cpp_s2 = GaussianAccumulatorS2(**kwargs_s2)

    query_max_phi = kwargs_base['max_phi'] - 5

    for i, (mesh_fpath, r) in enumerate(zip(ALL_MESHES, ALL_MESHES_ROTATIONS)):
        if i < 0:
            continue
        fname = mesh_fpath.stem
        # print(fname)
        example_mesh = o3d.io.read_triangle_mesh(str(mesh_fpath))
        example_mesh_filtered = example_mesh
        if r is not None:
            example_mesh_filtered = example_mesh_filtered.rotate(r.as_matrix())
            example_mesh_filtered = example_mesh_filtered.filter_smooth_laplacian(5)
        # example_mesh_filtered = example_mesh.filter_smooth_taubin(1)
        example_mesh_filtered.compute_triangle_normals()

        colored_icosahedron_py, normals, neighbors_py = visualize_gaussian_integration(
            ga_py_kdd, example_mesh_filtered, max_phi=query_max_phi)
        colored_icosahedron_cpp, normals, neighbors_cpp = visualize_gaussian_integration(
            ga_cpp_kdd, example_mesh_filtered, max_phi=query_max_phi)
        colored_icosahedron_opt, normals, neighbors_opt = visualize_gaussian_integration(
            ga_cpp_opt, example_mesh_filtered, max_phi=query_max_phi, integrate_kwargs=kwargs_opt_integrate)
        colored_icosahedron_s2, normals, neighbors_s2 = visualize_gaussian_integration(
            ga_cpp_s2, example_mesh_filtered, max_phi=query_max_phi, integrate_kwargs=kwargs_opt_integrate)

        idx_opt, = np.nonzero(neighbors_opt.astype(np.int64) - neighbors_cpp.astype(np.int64))
        print("Fast GaussianAccumulatorOpt (Hemisphere) incorrectly assigned : {}".format(idx_opt.shape[0]))

        reorder_s2 = reorder_single_hv_to_s2(ga_cpp_kdd, ga_cpp_s2)
        idx_s2, = np.nonzero(reorder_s2[neighbors_cpp].astype(np.int64) - neighbors_s2.astype(np.int64))
        print("Fast GaussianAccumulatorS2 (Full Sphere) incorrectly assigned : {}".format(
            idx_s2.shape[0]))  # Doesn't work because sorting is different

        if idx_opt.shape[0] > 0:
            pass
            # plot_issues_2(idx_opt, normals, neighbors_opt, ga_cpp_opt, colored_icosahedron_opt)

        plot_meshes(colored_icosahedron_s2, example_mesh_filtered)
        # plot_meshes(colored_icosahedron_py, colored_icosahedron_cpp,
        #             colored_icosahedron_opt, colored_icosahedron_s2, example_mesh_filtered)
        # plot_meshes(colored_icosahedron_s2, example_mesh_filtered)
        # plot_hilbert_curve(ga_py_kdd)
        # normals_sorted = plot_hilbert_curve(ga_cpp_kdd)
        pcd_cpp_opt = plot_hilbert_curve(ga_cpp_opt, plot=False)
        pcd_cpp_s2 = plot_hilbert_curve(ga_cpp_s2, plot=False)
        pcd_cpp_s2_image = get_image_peaks(ga_cpp_s2, **kwargs_base)

        normals_sorted_proj_hilbert = np.asarray(ga_cpp_opt.get_bucket_normals())
        normals_sorted_cube_hilbert = np.asarray(ga_cpp_s2.get_bucket_normals())
        plot_meshes([colored_icosahedron_s2, create_line_set(normals_sorted_cube_hilbert * 1.01), *pcd_cpp_s2],
                    [colored_icosahedron_s2, create_line_set(normals_sorted_cube_hilbert * 1.01), *pcd_cpp_s2_image])
        # plot_meshes([colored_icosahedron_cpp, create_line_set(normals_sorted_proj_hilbert * 1.01), *pcd_cpp_opt],
        #             [colored_icosahedron_s2, create_line_set(normals_sorted_cube_hilbert * 1.01), *pcd_cpp_s2],
        #             [colored_icosahedron_s2, create_line_set(normals_sorted_cube_hilbert * 1.01), *pcd_cpp_s2_image])

        ga_py_kdd.clear_count()
        ga_cpp_kdd.clear_count()
        ga_cpp_opt.clear_count()
        ga_cpp_s2.clear_count()


if __name__ == "__main__":
    main()



import time
from pathlib import Path
from collections import namedtuple

import numpy as np
import open3d as o3d
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt

from fastgac import GaussianAccumulatorKD, GaussianAccumulatorOpt, GaussianAccumulatorS2, MatX3d, convert_normals_to_hilbert, IcoCharts, GaussianAccumulatorS2Beta
from fastgac.peak_and_cluster import find_peaks_from_accumulator, find_peaks_from_ico_charts
from fastgac.o3d_util import get_arrow, get_pc_all_peaks, get_arrow_normals

from examples.python.run_meshes import visualize_gaussian_integration, plot_meshes
from examples.python.util.mesh_util import ALL_MESHES, ALL_MESHES_ROTATIONS

def main():
    EXAMPLE_INDEX = 1
    kwargs_base = dict(level=4)
    kwargs_s2 = dict(**kwargs_base)
    kwargs_opt_integrate = dict(num_nbr=12)
    query_max_phi = 175

    # Get an Example Mesh
    ga_cpp_s2 = GaussianAccumulatorS2Beta(**kwargs_s2)

    example_mesh = o3d.io.read_triangle_mesh(str(ALL_MESHES[EXAMPLE_INDEX]))
    r = ALL_MESHES_ROTATIONS[EXAMPLE_INDEX]
    example_mesh_filtered = example_mesh
    if r is not None:
        example_mesh_filtered = example_mesh_filtered.rotate(r.as_matrix())
        example_mesh_filtered = example_mesh_filtered.filter_smooth_laplacian(5)

    example_mesh_filtered.compute_triangle_normals()
    # np.save('fixtures/normals/basement.npy', np.asarray(example_mesh_filtered.triangle_normals))
    colored_icosahedron_s2, normals, neighbors_s2 = visualize_gaussian_integration(
        ga_cpp_s2, example_mesh_filtered, max_phi=query_max_phi, integrate_kwargs=kwargs_opt_integrate)


    o3d.visualization.draw_geometries([example_mesh_filtered])
    o3d.visualization.draw_geometries([colored_icosahedron_s2])


    # Visualize unwrapping
    ico_chart_ = IcoCharts(kwargs_base['level'])
    t2 = time.perf_counter()
    normalized_bucket_counts_by_vertex = ga_cpp_s2.get_normalized_bucket_counts_by_vertex(True)
    ico_chart_.fill_image(normalized_bucket_counts_by_vertex)

    average_bucket_normals = np.asarray(ga_cpp_s2.get_bucket_average_normals(True))
    pcd = o3d.geometry.PointCloud(o3d.utility.Vector3dVector(average_bucket_normals))
    pcd.paint_uniform_color([1, 0, 0])
    average_vertex_normals = np.asarray(ga_cpp_s2.get_average_normals_by_vertex(True))


    find_peaks_kwargs=dict(threshold_abs=50,min_distance=1, exclude_border=False, indices=False)
    print(np.asarray(ico_chart_.image).shape)
    cluster_kwargs=dict(t =0.1,criterion ='distance')
    _, _, avg_peaks, avg_weights = find_peaks_from_ico_charts(ico_chart_, np.asarray(normalized_bucket_counts_by_vertex), vertices=average_vertex_normals, find_peaks_kwargs=find_peaks_kwargs, cluster_kwargs=cluster_kwargs)
    t3 = time.perf_counter()
    print(t3 -t2)
    print(avg_peaks)
    # import ipdb; ipdb.set_trace()



    arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)
    wireframe = o3d.geometry.LineSet.create_from_triangle_mesh(colored_icosahedron_s2)
    o3d.visualization.draw_geometries([colored_icosahedron_s2, *arrow_avg_peaks, wireframe])
    # o3d.visualization.draw_geometries([colored_icosahedron_s2, *arrow_avg_peaks, pcd])

    full_image = np.asarray(ico_chart_.image)

    plt.imshow(full_image)
    plt.axis('off')
    # plt.xticks(np.arange(0, full_image.shape[1], step=1))
    # plt.yticks(np.arange(0, full_image.shape[0], step=1))
    plt.show()



if __name__ == "__main__":
    main()


"""Mesh
{
	"class_name" : "ViewTrajectory",
	"interval" : 29,
	"is_loop" : false,
	"trajectory" : 
	[
		{
			"boundingbox_max" : [ 1.8764505760969685, 3.0280973667097442, 3.045776668203259 ],
			"boundingbox_min" : [ -2.2365574934452548, -3.6804227036671078, 0.51828136237409295 ],
			"field_of_view" : 60.0,
			"front" : [ -0.43966986583569911, 0.57136927624194478, 0.69298453030552898 ],
			"lookat" : [ 0.30001921841467899, -0.99779994278506134, 1.5071575255263165 ],
			"up" : [ 0.44135525764305411, -0.53453483690843095, 0.72074825333268089 ],
			"zoom" : 0.31999999999999978
		}
	],
	"version_major" : 1,
	"version_minor" : 0
}
"""


"""
{
	"class_name" : "ViewTrajectory",
	"interval" : 29,
	"is_loop" : false,
	"trajectory" : 
	[
		{
			"boundingbox_max" : [ 1.1339119391275889, 1.1343327326857235, 1.1998729449684717 ],
			"boundingbox_min" : [ -1.1353148374296551, -1.0, -1.1999606130137823 ],
			"field_of_view" : 60.0,
			"front" : [ -0.59564118276660283, 0.48513744010499366, 0.6401978175538996 ],
			"lookat" : 
			[
				-0.00070144915103309557,
				0.067166366342861772,
				-4.3834022655286908e-05
			],
			"up" : [ 0.47207151576167344, -0.43341779039025202, 0.76765715197587236 ],
			"zoom" : 0.69999999999999996
		}
	],
	"version_major" : 1,
	"version_minor" : 0
}
"""


"""This example is a simply demo with no visualization of integrating normals into level 0 GA
"""
import numpy as np
import time 

from fastgac import GaussianAccumulatorKD, MatX3d

np.set_printoptions(suppress=True, precision=2)

def main():
    normals = np.array([
        [-1, 0, 0],
        [0, 0, 1],
        [1, 0, 0],
        [0, 1, 0],
        [0, -1, 0]
    ])
    t0 = time.perf_counter()
    ga = GaussianAccumulatorKD(level=0, max_phi=180)
    t1 = time.perf_counter()
    print("Number of Buckets: {}\n".format(len(ga.buckets)))
    print("Bucket representations:\n {}\n".format(ga.buckets))
    print("Bucket Cell Surface Normals: \n {}\n".format(np.asarray(ga.get_bucket_normals())))
    normals_mat = MatX3d(normals) # need to convert a format we understand
    t2 = time.perf_counter()
    bucket_indexes = ga.integrate(normals_mat)
    t3 = time.perf_counter()
    print("These normals: \n {} \n are most similar to these cell normlas: \n {} \n".format(normals, np.asarray(ga.get_bucket_normals())[bucket_indexes,:]))
    print(np.asarray(bucket_indexes))
    print("Building Index Took (ms): {}; Query Took (ms): {}".format((t1-t0) * 1000, (t3 - t2)* 1000))

    print("Change the level see a better approximation")
    

if __name__ == "__main__":
    main()



import numpy as np
import open3d as o3d
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt
import matplotlib.cm as cm

from fastgac import GaussianAccumulatorKD, GaussianAccumulatorOpt, GaussianAccumulatorS2, MatX3d, convert_normals_to_hilbert, IcoCharts
from fastgac.peak_and_cluster import find_peaks_from_accumulator, find_peaks_from_ico_charts
from fastgac.o3d_util import get_arrow, get_pc_all_peaks, get_arrow_normals, create_open_3d_mesh, assign_vertex_colors, plot_meshes

from examples.python.run_meshes import create_line_set
from examples.python.util.line_mesh import LineMesh
# from src.Pyth import create_open_3d_mesh, assign_vertex_colors, plot_meshes

def main():
    kwargs_base = dict(level=4, max_phi=180)
    kwargs_s2 = dict(**kwargs_base)
    ga_cpp_s2 = GaussianAccumulatorS2(**kwargs_s2)
    normals_sorted_cube_hilbert = np.asarray(ga_cpp_s2.get_bucket_normals())

    refined_icosahedron_mesh = create_open_3d_mesh(np.asarray(ga_cpp_s2.mesh.triangles), np.asarray(ga_cpp_s2.mesh.vertices))

    indices = np.asarray(ga_cpp_s2.get_bucket_sfc_values())
    colors = indices / np.iinfo(indices.dtype).max
    colors = cm.viridis(colors)[:, :3]

    lm = LineMesh(normals_sorted_cube_hilbert * 1.01, colors=colors, radius=0.004)

    # colored_icosahedron = assign_vertex_colors(refined_icosahedron_mesh, colors)
    # plot_meshes([refined_icosahedron_mesh, create_line_set(normals_sorted_cube_hilbert * 1.01)])
    ls = o3d.geometry.LineSet.create_from_triangle_mesh(refined_icosahedron_mesh)
    refined_icosahedron_mesh.paint_uniform_color([0.5, 0.5, 0.5])
    plot_meshes([refined_icosahedron_mesh, ls, *lm.cylinder_segments])


if __name__ == "__main__":
    main()


# {
# 	"class_name" : "ViewTrajectory",
# 	"interval" : 29,
# 	"is_loop" : false,
# 	"trajectory" : 
# 	[
# 		{
# 			"boundingbox_max" : [ 1.7098574459629279, 0.70987207591299273, 1.0130857360820043 ],
# 			"boundingbox_min" : [ 0.29014255403707201, -1.0130857026947551, -1.0130857360820043 ],
# 			"field_of_view" : 60.0,
# 			"front" : [ -0.0040044546059786173, 0.11016896364951138, 0.99390480620213395 ],
# 			"lookat" : [ 1.0, -0.15160681339088117, 0.0 ],
# 			"up" : [ 0.0019505060267622787, 0.99391174507838853, -0.11016187417374566 ],
# 			"zoom" : 0.69999999999999996
# 		}
# 	],
# 	"version_major" : 1,
# 	"version_minor" : 0
# }


import time
from pathlib import Path
from collections import namedtuple

import numpy as np
import open3d as o3d
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt

from fastgac import GaussianAccumulatorKD, GaussianAccumulatorOpt, GaussianAccumulatorS2, MatX3d, convert_normals_to_hilbert, IcoCharts
from fastgac.peak_and_cluster import find_peaks_from_accumulator, find_peaks_from_ico_charts
from fastgac.o3d_util import get_arrow, get_pc_all_peaks, get_arrow_normals

from examples.python.run_meshes import visualize_gaussian_integration, plot_meshes
from examples.python.util.mesh_util import ALL_MESHES, ALL_MESHES_ROTATIONS

def main():
    EXAMPLE_INDEX = 1
    kwargs_base = dict(level=2, max_phi=180)
    kwargs_s2 = dict(**kwargs_base)
    kwargs_opt_integrate = dict(num_nbr=12)
    query_max_phi = kwargs_base['max_phi'] - 5

    # Get an Example Mesh
    ga_cpp_s2 = GaussianAccumulatorS2(**kwargs_s2)

    example_mesh = o3d.io.read_triangle_mesh(str(ALL_MESHES[EXAMPLE_INDEX]))
    r = ALL_MESHES_ROTATIONS[EXAMPLE_INDEX]
    example_mesh_filtered = example_mesh
    if r is not None:
        example_mesh_filtered = example_mesh_filtered.rotate(r.as_matrix())
        example_mesh_filtered = example_mesh_filtered.filter_smooth_laplacian(5)

    example_mesh_filtered.compute_triangle_normals()
    # np.save('fixtures/normals/basement.npy', np.asarray(example_mesh_filtered.triangle_normals))
    colored_icosahedron_s2, normals, neighbors_s2 = visualize_gaussian_integration(
        ga_cpp_s2, example_mesh_filtered, max_phi=query_max_phi, integrate_kwargs=kwargs_opt_integrate)


    o3d.visualization.draw_geometries([example_mesh_filtered])
    o3d.visualization.draw_geometries([colored_icosahedron_s2])


    # Visualize unwrapping
    ico_chart_ = IcoCharts(kwargs_base['level'])
    t2 = time.perf_counter()
    normalized_bucket_counts_by_vertex = ga_cpp_s2.get_normalized_bucket_counts_by_vertex(True)
    ico_chart_.fill_image(normalized_bucket_counts_by_vertex)

    find_peaks_kwargs=dict(threshold_abs=50,min_distance=1, exclude_border=False, indices=False)
    print(np.asarray(ico_chart_.image).shape)
    cluster_kwargs=dict(t =0.1,criterion ='distance')
    _, _, avg_peaks, avg_weights = find_peaks_from_ico_charts(ico_chart_, np.asarray(normalized_bucket_counts_by_vertex), find_peaks_kwargs=find_peaks_kwargs, cluster_kwargs=cluster_kwargs)
    t3 = time.perf_counter()
    print(t3 -t2)
    print(avg_peaks)

    arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)
    o3d.visualization.draw_geometries([colored_icosahedron_s2, *arrow_avg_peaks])

    full_image = np.asarray(ico_chart_.image)

    plt.imshow(full_image)
    plt.axis('off')
    # plt.xticks(np.arange(0, full_image.shape[1], step=1))
    # plt.yticks(np.arange(0, full_image.shape[0], step=1))
    plt.show()



if __name__ == "__main__":
    main()



import numpy as np
import open3d as o3d

GOLDEN_RATIO = (1.0 + np.sqrt(5.0)) / 2.0
ICOSAHEDRON_TRUE_RADIUS = np.sqrt(1 + np.power(GOLDEN_RATIO, 2))
ICOSAHEDRON_SCALING = 1.0 / ICOSAHEDRON_TRUE_RADIUS


def create_icosahedron(radius=ICOSAHEDRON_SCALING):
    if (radius <= 0):
        ValueError("Radius too small")

    p = GOLDEN_RATIO
    vertices = []
    triangles = []
    vertices.append(radius * np.array([-1, 0, p]))
    vertices.append(radius * np.array([1, 0, p]))
    vertices.append(radius * np.array([1, 0, -p]))
    vertices.append(radius * np.array([-1, 0, -p]))
    vertices.append(radius * np.array([0, -p, 1]))
    vertices.append(radius * np.array([0, p, 1]))
    vertices.append(radius * np.array([0, p, -1]))
    vertices.append(radius * np.array([0, -p, -1]))
    vertices.append(radius * np.array([-p, -1, 0]))
    vertices.append(radius * np.array([p, -1, 0]))
    vertices.append(radius * np.array([p, 1, 0]))
    vertices.append(radius * np.array([-p, 1, 0]))
    triangles.append(np.array([0, 4, 1]))
    triangles.append(np.array([0, 1, 5]))
    triangles.append(np.array([1, 4, 9]))
    triangles.append(np.array([1, 9, 10]))
    triangles.append(np.array([1, 10, 5]))
    triangles.append(np.array([0, 8, 4]))
    triangles.append(np.array([0, 11, 8]))
    triangles.append(np.array([0, 5, 11]))
    triangles.append(np.array([5, 6, 11]))
    triangles.append(np.array([5, 10, 6]))
    triangles.append(np.array([4, 8, 7]))
    triangles.append(np.array([4, 7, 9]))
    triangles.append(np.array([3, 6, 2]))
    triangles.append(np.array([3, 2, 7]))
    triangles.append(np.array([2, 6, 10]))
    triangles.append(np.array([2, 10, 9]))
    triangles.append(np.array([2, 9, 7]))
    triangles.append(np.array([3, 11, 6]))
    triangles.append(np.array([3, 8, 11]))
    triangles.append(np.array([3, 7, 8]))
    triangles = np.array(triangles)
    vertices = np.array(vertices)
    return triangles, vertices


def create_icosahedron_mesh(radius=ICOSAHEDRON_SCALING):
    triangles, vertices = create_icosahedron(radius=radius)
    # mesh = o3d.geometry.TriangleMesh(o3d.utility.Vector3dVector(vertices), o3d.utility.Vector3iVector(triangles))
    mesh = o3d.geometry.TriangleMesh.create_icosahedron(radius=ICOSAHEDRON_SCALING)
    return mesh


def cantor_mapping(k1, k2):
    return int(((k1 + k2) * (k1 + k2 + 1)) / 2.0 + k2)


def generate_key_from_point(p1_idx, p2_idx):
    lower_idx, higher_idx = (
        p1_idx, p2_idx) if p1_idx < p2_idx else (p2_idx, p1_idx)
    return cantor_mapping(lower_idx, higher_idx)


def construct_midpoint(p1_idx, p2_idx, vertices):
    p1 = vertices[p1_idx]
    p2 = vertices[p2_idx]
    midpoint_on_plane = (p2 + p1) / 2.0
    scaling = 1 / np.linalg.norm(midpoint_on_plane)
    midpoint_on_sphere = midpoint_on_plane * scaling
    return midpoint_on_sphere


def get_point_idx(p1_idx, p2_idx, point_to_idx_map, vertices):
    point_key = generate_key_from_point(p1_idx, p2_idx)
    if point_to_idx_map.get(point_key):
        # Existing point Idx
        return point_to_idx_map[point_key]
    else:
        # New point idx
        point_to_idx_map[point_key] = len(vertices)
        midpoint_on_sphere = construct_midpoint(p1_idx, p2_idx, vertices)
        vertices.append(midpoint_on_sphere)
        return point_to_idx_map[point_key]


def refine_icosahedron(triangles, vertices, level=2):
    vertices = list(vertices)
    triangles = triangles.tolist()
    # HashMap that maps a midpoint (identified by two point indexes) to its own point index
    point_to_idx_map = dict()
    for i in range(level):
        triangles_refined = []
        # loop through every triangle and create 4 new ones based upon midpoints
        for triangle in triangles:
            p1_idx = triangle[0]
            p2_idx = triangle[1]
            p3_idx = triangle[2]

            # Create new points (if not existing) and return point index
            p4_idx = get_point_idx(p1_idx, p2_idx, point_to_idx_map, vertices)
            p5_idx = get_point_idx(p2_idx, p3_idx, point_to_idx_map, vertices)
            p6_idx = get_point_idx(p3_idx, p1_idx, point_to_idx_map, vertices)
            # Create the four new triangles
            t1 = [p1_idx, p4_idx, p6_idx]
            t2 = [p4_idx, p2_idx, p5_idx]
            t3 = [p6_idx, p5_idx, p3_idx]
            t4 = [p6_idx, p4_idx, p5_idx]
            # Append triangles to the new refined triangle array
            triangles_refined.extend([t1, t2, t3, t4])
        # overwrite existing triangles with this new array
        triangles = triangles_refined
    vertices = np.array(vertices)
    triangles = np.array(triangles)
    return vertices, triangles



import time
from pathlib import Path
from collections import namedtuple
import open3d as o3d
import numpy as np
from scipy.spatial import cKDTree, KDTree
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt
from hilbertcurve.hilbertcurve import HilbertCurve

from .helper import assign_vertex_colors, create_open_3d_mesh, calc_angle_delta, plot_meshes, draw_normals
from .icosahedron import create_icosahedron, refine_icosahedron, create_icosahedron_mesh
from .projections import plot_projection, convert_lat_long, azimuth_equidistant
from .peak_and_cluster import find_peaks_from_accumulator

THIS_DIR = Path(__file__).parent
FIXTURES_DIR = THIS_DIR / "../../../fixtures/"
REALSENSE_DIR = (FIXTURES_DIR / "realsense").absolute()
EXAMPLE_MESH_1 = REALSENSE_DIR / "example_mesh.ply"
EXAMPLE_MESH_2 = REALSENSE_DIR / "dense_first_floor_map.ply"
EXAMPLE_MESH_3 = REALSENSE_DIR / "sparse_basement.ply"

ALL_MESHES = [EXAMPLE_MESH_1, EXAMPLE_MESH_2, EXAMPLE_MESH_3]
ALL_MESHES_ROTATIONS = [None, R.from_rotvec(-np.pi / 2 * np.array([1, 0, 0])),
                        R.from_rotvec(-np.pi / 2 * np.array([1, 0, 0]))]

def get_colors(inp, colormap=plt.cm.viridis, vmin=None, vmax=None):
    norm = plt.Normalize(vmin, vmax)
    return colormap(norm(inp))


def refined_ico_mesh(level=0):
    triangles, vertices = create_icosahedron()
    vertices, triangles = refine_icosahedron(triangles, vertices, level=level)
    new_mesh = create_open_3d_mesh(triangles, vertices)
    return np.array(new_mesh.triangle_normals), new_mesh


class GaussianAccumulatorKDPy(object):
    def __init__(self, level=0, max_phi=100, max_leaf_size=16):
        super().__init__()
        bucket_normals, mesh = refined_ico_mesh(level=level)
        bucket_normals, mask = filter_normals_by_phi(bucket_normals, max_phi)
        self.bucket_normals = bucket_normals
        self.bucket_indices = assign_hilbert_curve(bucket_normals)
        self.mask = mask
        self.mesh = mesh
        self.num_buckets = bucket_normals.shape[0]
        self.accumulator = np.zeros(self.num_buckets, dtype=np.int32)
        self.accumulator_normalized = np.zeros(self.num_buckets, dtype=np.float64)
        self.bucket_projection = azimuth_equidistant(bucket_normals)

        # Sort buckets by hilbert indices
        idx_sort = np.argsort(self.bucket_indices)
        self.bucket_indices  = np.ascontiguousarray(self.bucket_indices[idx_sort])
        self.bucket_projection  = np.ascontiguousarray(self.bucket_projection[idx_sort])
        self.bucket_normals  = np.ascontiguousarray(self.bucket_normals[idx_sort])
        # fix triangles
        triangles = np.asarray(self.mesh.triangles)
        vertices = np.asarray(self.mesh.vertices)
        triangles = np.vstack((triangles[mask][idx_sort], triangles[~mask]))
        self.mesh = create_open_3d_mesh(triangles, vertices)
        self.kdtree = cKDTree(self.bucket_normals, leafsize=max_leaf_size, compact_nodes=True, balanced_tree=True)

    def integrate(self, normals):
        _, neighbors = self.kdtree.query(normals)
        np.add.at(self.accumulator, neighbors, 1)
        return neighbors
        # for idx in neighbors:
        #     self.accumulator[idx] = self.accumulator[idx] + 1
    
    def get_bucket_normals(self):
        return self.bucket_normals

    def get_normalized_bucket_counts(self):
        # print(np.max(self.accumulator))
        self.accumulator_normalized = self.accumulator / np.max(self.accumulator)
        return self.accumulator_normalized

    def get_bucket_sfc_values(self):
        return self.bucket_indices

    def get_bucket_projection(self):
        return self.bucket_projection

    def clear_count(self):
        self.accumulator.fill(0)


class GaussianAccumulator(object):
    def __init__(self, gaussian_normals, leafsize=16):
        super().__init__()
        self.nbuckets = gaussian_normals.shape[0]
        self.kdtree = cKDTree(gaussian_normals, leafsize=leafsize, compact_nodes=True, balanced_tree=True)
        self.accumulator = np.zeros(self.nbuckets, dtype=np.int32)
        self.accumulator_normalized = np.zeros(self.nbuckets, dtype=np.float64)
        self.colors = np.zeros_like(gaussian_normals)
        self.gaussian_normals = gaussian_normals
        self.normals = None

    def integrate(self, normals):
        self.normals = normals
        query_size = normals.shape[0]
        t0 = time.perf_counter()
        _, neighbors = self.kdtree.query(normals)
        t1 = time.perf_counter()
        elapsed_time = (t1 - t0) * 1000
        print("KD tree size: {}; Query Size (K): {}; Execution Time(ms): {:.1f}".format(
            self.nbuckets, query_size, elapsed_time))
        for idx in neighbors:
            self.accumulator[idx] = self.accumulator[idx] + 1

    def normalize(self):
        self.accumulator_normalized = self.accumulator / np.max(self.accumulator)
        self.colors = get_colors(self.accumulator, plt.cm.viridis)[:, :3]


def generate_family_of_icosahedron(triangles, vertices, family=[1, 2, 3, 4]):
    meshes = []
    for level in family:
        new_vertices, new_triangles = refine_icosahedron(
            triangles, vertices, level=level)
        new_mesh = create_open_3d_mesh(new_triangles, new_vertices)
        meshes.append(new_mesh)
    return meshes


def generate_sphere_examples():
    ico = create_icosahedron_mesh()
    sphere = o3d.geometry.TriangleMesh.create_sphere(resolution=20)

    ico.compute_vertex_normals()
    ico.compute_triangle_normals()
    sphere.compute_vertex_normals()
    sphere.compute_triangle_normals()

    return ico, sphere


def visualize_refinement(ico, level=2, plot=False):
    vertices, triangles = refine_icosahedron(np.asarray(
        ico.triangles), np.asarray(ico.vertices), level=level)
    # print(vertices, vertices.shape)
    # print(triangles, triangles.shape)
    new_mesh = create_open_3d_mesh(triangles, vertices)
    # create lineset of normals
    top_normals = np.asarray(new_mesh.triangle_normals)
    # mask = top_normals[:, 2] >= 0.0
    # new_mesh.remove_triangles_by_mask(~mask)
    # top_normals = np.ascontiguousarray(top_normals[top_normals[:, 2] >= 0.0, :])

    line_set = draw_normals(top_normals)
    pcd_normals = o3d.geometry.PointCloud(
        o3d.utility.Vector3dVector(top_normals))
    pcd_normals.paint_uniform_color([0.5, 0.5, 0.5])
    if plot:
        plot_meshes(new_mesh, line_set, pcd_normals, shift=False)
    return new_mesh, top_normals


def filter_normals_by_phi(normals, max_phi=100, return_mask=True):
    rad = np.deg2rad(max_phi)
    # TODO this can be reduced to simple z comparison
    min_z = np.cos(rad)
    # phi = np.arctan2(1.0, normals[:, 2])
    # mask = np.abs(phi) < rad
    mask = normals[:, 2] >= min_z
    if return_mask:
        return normals[mask, :], mask
    else:
        return normals[mask, :]


# def assign_hilbert_curve(ga: GaussianAccumulator):
#     dtype = np.uint32
#     max_length_axis = 2**16 - 1
#     bucket_normals = ga.gaussian_normals
#     bucket_normals_xy = (azimuth_equidistant(bucket_normals) * max_length_axis).astype(dtype)
#     hilbert_curve = HilbertCurve(16, 2)
#     bucket_normals_hv = []
#     for i in range(bucket_normals_xy.shape[0]):
#         hv = hilbert_curve.distance_from_coordinates(bucket_normals_xy[i, :])
#         bucket_normals_hv.append(hv)
#     bucket_normals_hv = np.array(bucket_normals_hv)
#     return bucket_normals_hv

def assign_hilbert_curve(normals:np.ndarray):
    dtype = np.uint32
    max_length_axis = 2**16 - 1
    bucket_normals = normals
    bucket_normals_xy = (azimuth_equidistant(bucket_normals) * max_length_axis).astype(dtype)
    # print(bucket_normals_xy)
    hilbert_curve = HilbertCurve(16, 2)
    bucket_normals_hv = []
    for i in range(bucket_normals_xy.shape[0]):
        hv = hilbert_curve.distance_from_coordinates(bucket_normals_xy[i, :])
        bucket_normals_hv.append(hv)
    bucket_normals_hv = np.array(bucket_normals_hv)
    return bucket_normals_hv


def plot_hilbert_curve(ga):
    bucket_normals = ga.gaussian_normals
    colors = ga.colors
    proj = azimuth_equidistant(bucket_normals)
    bucket_normals_hv = assign_hilbert_curve(ga.gaussian_normals)
    # print(np.max(bucket_normals_hv), np.min(bucket_normals_hv))
    idx_sort = np.argsort(bucket_normals_hv)
    proj = proj[idx_sort, :]
    accumulator_normalized_sorted = ga.accumulator_normalized[idx_sort]
    gaussian_normals_sorted = ga.gaussian_normals[idx_sort]

    # Find Peaks
    peaks, clusters, _, _ = find_peaks_from_accumulator(gaussian_normals_sorted, accumulator_normalized_sorted)

    colors = colors[idx_sort, :]
    fig, axs = plt.subplots(2, 1, figsize=(8, 10))
    ax = axs[0]
    scatter1 = ax.scatter(proj[:, 0], proj[:, 1], c=colors, label='Projected Buckets')
    scatter2 = ax.scatter(proj[peaks, :][:, 0], proj[peaks, :][:, 1], marker='x', c=clusters, label='Clusters', cmap='tab20')
    ax.set_title("Hilbert Curve with Azimuth Equidistant Projection")
    ax.set_xlabel("x*")
    ax.set_ylabel("y*")
    line1 = ax.plot(proj[:, 0], proj[:, 1], c='k', label='Hilbert Curve Connections')[0]
    ax.axis('equal')
    leg = ax.legend(loc='upper left', fancybox=True, shadow=True)

    # we will set up a dict mapping legend line to orig line, and enable
    # picking on the legend line
    lines = [line1, scatter1, scatter2]
    lined = dict()
    for legline, origline in zip(leg.legendHandles, lines):
        legline.set_picker(5)  # 5 pts tolerance
        lined[legline] = origline

    def onpick(event):
        # on the pick event, find the orig line corresponding to the
        # legend proxy line, and toggle the visibility
        legline = event.artist
        origline = lined[legline]
        vis = not origline.get_visible()
        origline.set_visible(vis)
        # Change the alpha on the line in the legend so we can see what lines
        # have been toggled
        if vis:
            legline.set_alpha(1.0)
        else:
            legline.set_alpha(0.2)
        fig.canvas.draw()

    ax = axs[1]
    ax.bar(np.arange(ga.nbuckets), accumulator_normalized_sorted)
    ax.scatter(peaks, accumulator_normalized_sorted[peaks], marker='x', c=clusters, cmap='tab20')

    ax.set_title("Histogram of Normal Counts sorted by Hilbert Values")
    ax.set_xlabel("Hilbert Value (Ascending)")
    ax.set_ylabel("Normal Counts")

    # ax.axis('equal')
    fig.canvas.mpl_connect('pick_event', onpick)
    fig.tight_layout()
    plt.show()

# def find_peaks_from_accumulator(gaussian_normals_sorted, accumulator_normalized_sorted,
#                find_peaks_kwargs=dict(height=0.10, threshold=None, distance=4, width=None, prominence=None),
#                cluster_kwargs=dict(t=0.15, criterion='distance')):
#     t0 = time.perf_counter()
#     peaks, _ = find_peaks(accumulator_normalized_sorted, **find_peaks_kwargs)
#     t1 = time.perf_counter()

#     gaussian_normal_1d_clusters = gaussian_normals_sorted[peaks,:]
#     Z = linkage(gaussian_normal_1d_clusters, 'single')
#     clusters = fcluster(Z, **cluster_kwargs)
#     t2 = time.perf_counter()

#     print("Peak Detection - Find Peaks Execution Time (ms): {:.1f}; Hierarchical Clustering Execution Time (ms): {:.1f}".format((t1-t0) * 1000, (t2-t1) * 1000))
#     return peaks, clusters


def visualize_gaussian_integration(refined_icosahedron_mesh, gaussian_normals, mesh, ds=50, min_samples=10000, plot=False):
    to_integrate_normals = np.asarray(mesh.triangle_normals)
    # remove normals on bottom half of sphere
    to_integrate_normals, _ = filter_normals_by_phi(to_integrate_normals)
    # determine optimal sampling
    num_normals = to_integrate_normals.shape[0]
    ds_normals = int(num_normals / ds)
    to_sample = max(min([num_normals, min_samples]), ds_normals)
    # perform sampling of normals
    to_integrate_normals = to_integrate_normals[np.random.choice(
        num_normals, to_sample), :]
    # integrate normals
    gaussian_normals, mask = filter_normals_by_phi(gaussian_normals, return_mask=True)
    ga = GaussianAccumulator(gaussian_normals)
    ga.integrate(to_integrate_normals)
    ga.normalize()
    # Colorize normal buckets
    colored_icosahedron = assign_vertex_colors(
        refined_icosahedron_mesh, ga.colors, mask)
    if plot:
        plot_meshes(colored_icosahedron, mesh)

    return ga, colored_icosahedron


def main():

    ico, sphere = generate_sphere_examples()
    ico_copy = o3d.geometry.TriangleMesh(ico)
    family = [1, 2, 3, 4]
    meshes = generate_family_of_icosahedron(np.asarray(
        ico.triangles), np.asarray(ico.vertices), family)
    family.insert(0, 0)
    meshes.insert(0, ico)
    for level, mesh in zip(family, meshes):
        angle_diff = calc_angle_delta(mesh, level)
        print("Refinement Level: {}; Number of Vertices: {}, Number of Triangles: {}, Angle Difference: {:.1f}".format(
            level, np.array(mesh.vertices).shape[0], np.array(mesh.triangles).shape[0], angle_diff))
    # meshes.insert(0, sphere)
    plot_meshes(*meshes)
    return
    # Show our chosen refined example
    refined_icosphere, gaussian_normals = visualize_refinement(
        ico_copy, level=4, plot=False)
    # print(gaussian_normals)
    # Get an Example Mesh
    for i, (mesh_fpath, r) in enumerate(zip(ALL_MESHES, ALL_MESHES_ROTATIONS)):
        if i < 0:
            continue
        fname = mesh_fpath.stem
        # print(fname)
        example_mesh = o3d.io.read_triangle_mesh(str(mesh_fpath))
        if r is not None:
            example_mesh = example_mesh.rotate(r.as_matrix())
        example_mesh.compute_triangle_normals()
        # plot_meshes(example_mesh)

        ga, colored_icosahedron = visualize_gaussian_integration(
            refined_icosphere, gaussian_normals, example_mesh)
        # np.savetxt(str(REALSENSE_DIR / "{}_buckets.txt".format(fname)),
        #            convert_lat_long(ga.gaussian_normals, degrees=True), fmt='%.4f')
        # np.savetxt(str(REALSENSE_DIR / "{}_normals.txt".format(fname)),
        #            convert_lat_long(ga.normals, degrees=True), fmt='%.4f')
        plot_projection(ga)
        plot_hilbert_curve(ga)
        plot_meshes(colored_icosahedron, example_mesh)


if __name__ == "__main__":
    main()


# 'filter_smooth_laplacian', 'filter_smooth_simple',
# for smooth_alg in ['filter_smooth_taubin']:
#     new_mesh = o3d.geometry.TriangleMesh(example_mesh)
#     func = getattr(new_mesh, smooth_alg)
#     new_mesh = func()
#     plot_meshes(new_mesh)
# Visualize Guassiant Integration



import numpy as np

def down_sample_normals(triangle_normals, down_sample_fraction=0.12, min_samples=10000, flip_normals=False, **kwargs):
    """Return dowsampled normals

    Args:
        triangle_normals (np.ndarray): Triangle Normals
        down_sample_fraction (float, optional): Fraction to downsample. Defaults to 0.12.
        min_samples (int, optional): Minimum number of samples. Defaults to 10000.
        flip_normals (bool, optional): Reverse the normals?. Defaults to False.

    Returns:
        np.ndarray: NX3 downsampled normals
    """
    num_normals = triangle_normals.shape[0]
    to_sample = int(down_sample_fraction * num_normals)
    to_sample = max(min([num_normals, min_samples]), to_sample)
    ds_step = int(num_normals / to_sample)
    triangle_normals_ds = np.ascontiguousarray(triangle_normals[:num_normals:ds_step, :])
    if flip_normals:
        triangle_normals_ds = triangle_normals_ds * -1.0
    return triangle_normals_ds


import numpy as np
import matplotlib.pyplot as plt
import time
from .helper import normalize_box, normalized

def convert_phi_theta(normals, top_half=True):
    phi_theta = np.zeros((normals.shape[0], 2))
    xy = normals[:, 0]**2 + normals[:, 1]**2
    phi_theta[:, 0] = np.arctan2(np.sqrt(xy), normals[:, 2])  # for elevation angle defined from Z-axis down
    phi_theta[:, 1] = np.arctan2(normals[:, 1], normals[:, 0])
    return phi_theta


def convert_lat_long(normals, degrees=True, return_mask=True):
    normals_new = np.copy(normals)
    phi_theta = np.zeros((normals_new.shape[0], 2))
    phi_theta[:, 0] = np.arccos(normals_new[:, 2])  # for elevation angle defined from Z-axis down
    phi_theta[:, 1] = np.arctan2(normals_new[:, 1], normals_new[:, 0])
    phi_theta[:, 0] = np.ones_like(phi_theta[:, 0]) * np.pi/2.0 - phi_theta[:, 0]
    if degrees:
        phi_theta[:, 0] = np.rad2deg(phi_theta[:, 0])
        phi_theta[:, 1] = np.rad2deg(phi_theta[:, 1])

    return phi_theta

def convert_stereographic(normals, top_half=True):
    normals_new = np.copy(normals)
    projection = np.zeros((normals_new.shape[0], 2))
    projection[:, 0] = normals_new[:, 0] / (1 - normals_new[:, 2])
    projection[:, 1] = normals_new[:, 1] / (1 - normals_new[:, 2])
    return projection


def convert_phi_theta_centered(normals, top_half=True):
    normals_new = np.copy(normals)
    projection = np.zeros((normals_new.shape[0], 2))
    xy = normals_new[:, 0]**2 + normals_new[:, 1]**2
    for i in range(normals_new.shape[0]):
        normal = normals_new[i, :]
        phi = np.arccos(normal[2])
        phi = -phi if normal[1] < 0 else phi
        theta = np.arctan2(normal[1], normal[0])
        theta = theta + np.pi / 2.0
        theta = theta - np.pi if theta > np.pi / 2.0 else theta
        projection[i, :] = [phi, theta]

    return projection


def down_proj(normals, top_half=True):
    normals_new = np.copy(normals)
    projection = np.zeros((normals_new.shape[0], 2))
    projection = normals_new[:, :2]

    return projection

def azimuth_equidistant(normals):
    projection = np.zeros((normals.shape[0], 2))
    theta = np.arctan2(normals[:, 1], normals[:, 0])
    phi = np.arccos(normals[:, 2])

    projection[:, 0] = phi * np.sin(theta)
    projection[:, 1] = - phi * np.cos(theta)
    # Normalize to 0-1
    projection = normalize_box(projection)

    return projection

def azimuth_equidistant_fast(normals):
    projection = np.zeros((normals.shape[0], 2))

    l2 = np.atleast_1d(np.linalg.norm(normals[:, :2], 2, -1))
    scaling = 1.0 / l2
    phi = np.arccos(normals[:, 2])

    projection[:, 0] = phi * normals[:, 1] * scaling
    projection[:, 1] =  -phi * normals[:, 0] * scaling
    # Normalize to 0-1
    projection = normalize_box(projection)

    return projection


def plot_projection(ga):

    projections = [("Spherical Coordinates", "phi", "theta", "convert_phi_theta"),
                   ("Spherical Coordinates Centered", "phi", "theta", "convert_phi_theta_centered"),
                   ("Geodetic", "lat", "lon", "convert_lat_long"),
                #    ("Steographic Projection", "x*", "y*", "convert_stereographic"),
                   ("Project To Plane", "x", "y", "down_proj"),
                   ("Azimuth Equidistant", "x*", "y*", "azimuth_equidistant"),
                   ("Azimuth Equidistant (Fast)", "x", "y", "azimuth_equidistant_fast"),
                   ]
    fig, axs = plt.subplots(3, 2, figsize=(5, 6))
    axs = axs.reshape(-1)
    for i, (title_name, xlabel, ylabel, function_name) in enumerate(projections):
        ax = axs[i]
        t0 = time.perf_counter()
        proj = globals()[function_name](ga.gaussian_normals)
        t1 = time.perf_counter()
        ax.scatter(proj[:, 0], proj[:, 1], c=ga.colors)
        ax.set_title(title_name)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.axis('equal')
        # print("{} took {}".format(title_name, t1-t0 ))
    fig.tight_layout()

    plt.show()


import time
import numpy as np
from scipy.signal import find_peaks
from scipy.cluster.hierarchy import linkage, fcluster
# from skimage.feature import peak_local_max
from fastgac.scikit_image.skimage_feature_peak import peak_local_max


def normalized(a, axis=-1, order=2):
    """Normalizes a numpy array of points"""
    l2 = np.atleast_1d(np.linalg.norm(a, order, axis))
    l2[l2 == 0] = 1
    return a / np.expand_dims(l2, axis), l2


def find_peaks_from_accumulator(gaussian_normals_sorted, accumulator_normalized_sorted,
                                find_peaks_kwargs=dict(height=0.05, threshold=None, distance=4,
                                                       width=None, prominence=0.07),
                                cluster_kwargs=dict(t=0.15, criterion='distance'),
                                average_filter=dict(min_total_weight=0.2)):
    """Used 1D peak detection on the Gaussian Accumulator sorted by Space Filling Curve index (hilbert curve or S2ID).

    By nature of this being a 1D signal of a 2D Manifold in 3D, multiple peaks close in 2D space will be detected.

    Group these detected peaks using Agglomerative Hierarchal Clustering (AHC), 

    After things are grouped, take weighted average of the group and filter any that don't meet a criteria.

    Args:
        gaussian_normals_sorted (np.ndarray): NX3 Array of the unit normals
        accumulator_normalized_sorted (np.ndarray): NX1 array of the normalized counts for each unit normal
        find_peaks_kwargs (dict, optional): 1D Signal Detector kwargs, see `scipy.signal.find_peaks`. Defaults to dict(height=0.05, threshold=None, distance=4, width=None, prominence=0.07).
        cluster_kwargs (dict, optional): AHC cluster kwargs. See `scipy.spatial.fcluster` Defaults to dict(t=0.15, criterion='distance').
        average_filter (dict, optional): Each group must have this much normalized value at min. Defaults to dict(min_total_weight=0.2).

    Returns:
        (np.ndarray, np.ndarray, np.ndarray, np.ndarray): integer index of peaks in `gaussian_normals_sorted`, interger groups by AHC, detected peak normals, the weights of the peaks
    """
    peaks, _ = find_peaks(accumulator_normalized_sorted, **find_peaks_kwargs)
    gaussian_normal_1d_clusters = gaussian_normals_sorted[peaks, :]
    if gaussian_normal_1d_clusters.shape[0] > 1:
        Z = linkage(gaussian_normal_1d_clusters, 'single')
        clusters = fcluster(Z, **cluster_kwargs)
    else:
        clusters = np.array([1])
    weights_1d_clusters = accumulator_normalized_sorted[peaks]
    average_peaks, average_weights = average_clusters(gaussian_normal_1d_clusters, weights_1d_clusters, clusters)
    return peaks, clusters, average_peaks, average_weights


def get_high_intensity_peaks(image, mask, num_peaks=np.inf):
    """
    Return the highest intensity peak coordinates.
    """
    # get coordinates of peaks
    coord = np.nonzero(mask)
    coord = np.column_stack(coord)
    # Highest peak first
    return coord[::-1]


def find_peaks_from_ico_charts(ico_charts, normalized_bucket_counts_by_vertex,
                               vertices=None,
                               find_peaks_kwargs=dict(threshold_abs=25, min_distance=1,
                                                      exclude_border=False, indices=False),
                               cluster_kwargs=dict(t=0.10, criterion='distance'),
                               average_filter=dict(min_total_weight=0.15)):
    """Detect peaks inside a 2D image from an unwrapped refined icosahedron.

    2D peak detection is carried out using skimage.feature.peak_local_max. Note the image inside ico_charts is already normalized between 0-255 (NOT 0-1), `find_peaks_kwargs`.

    After the peaks are detected we group them using Agglomerative Hierarchal Clustering (AHC), `cluster_kwargs`.

    After things are grouped, take weighted average of the group (using normalized_bucket_counts_by_vertex) and filter any that don't meet a min value, `average_filter`.

    Args:
        ico_charts (IcoChart): The IcoChart of the unwrapped refined icosahedron (GA)
        normalized_bucket_counts_by_vertex (np.ndarray): NX1 Array. The normalized histogram count (0-1) of the vertices of the icosahedron (used for to weight groups after AHC)
        find_peaks_kwargs (dict, optional): Keyword arguments for peak detection. See `skimage.feature.peak_local_max`. Defaults to dict(threshold_abs=25, min_distance=1, exclude_border=False, indices=False).
        cluster_kwargs (dict, optional): AHC cluster kwargs. See `scipy.spatial.fcluster` Defaults to dict(t=0.10, criterion='distance').
        average_filter ([type], optional): Each group must have this much normalized value at min. Defaults to dict(min_total_weight=0.15).

    Returns:
        [np.ndarray, np.ndarray, np.ndarray, np.ndarray]: integer index of peaks in `gaussian_normals_sorted`, interger groups by AHC, **detected peak normals**, the weights of the peaks
    """
    # Get data from ico chart
    # t0 = time.perf_counter()
    image_to_vertex_idx = np.asarray(ico_charts.image_to_vertex_idx)
    image = np.asarray(ico_charts.image)
    mask = np.asarray(ico_charts.mask)
    mask = np.ma.make_mask(mask, copy=False)
    vertices_mesh = np.asarray(ico_charts.sphere_mesh.vertices) if vertices is None else np.asarray(vertices)
    # 2D Peak Detection
    peak_mask = peak_local_max(image, **find_peaks_kwargs)
    # Filter out invalid peaks, get associated normals
    valid_peaks = mask & peak_mask
    peak_image_idx = get_high_intensity_peaks(image, valid_peaks)
    vertices_idx = image_to_vertex_idx[peak_image_idx[:, 0], peak_image_idx[:, 1]]
    unclustered_peak_normals = vertices_mesh[vertices_idx, :]
    # t1 = time.perf_counter()
    # print(peak_image_idx)
    if unclustered_peak_normals.shape[0] > 1:
        Z = linkage(unclustered_peak_normals, 'single')
        clusters = fcluster(Z, **cluster_kwargs)
    else:
        clusters = np.array([1])

    weights_1d_clusters = normalized_bucket_counts_by_vertex[vertices_idx]
    average_peaks, average_weights = average_clusters(
        unclustered_peak_normals, weights_1d_clusters, clusters, average_filter=average_filter)

    # print("IcoChart Peak Detection - Find Peaks Execution Time (ms): {:.1f}; Hierarchical Clustering Execution Time (ms): {:.1f}".format((t1-t0) * 1000, (t2-t1) * 1000))
    return vertices_idx, clusters, average_peaks, average_weights


def get_point_clusters(points, point_weights, clusters):
    """Cluster points (R3) together given a cluster grouping"""
    point_clusters = []
    cluster_groups = np.unique(clusters)
    for cluster in cluster_groups:
        temp_mask = clusters == cluster
        point_clusters.append((points[temp_mask, :], point_weights[temp_mask]))
    return point_clusters


def average_clusters(peaks, peak_weights, clusters, average_filter=dict(min_total_weight=0.15)):
    """Average any clusters together by weights, remove any that don't meet a minimum requirements"""
    cluster_points = get_point_clusters(peaks, peak_weights, clusters)
    clusters_averaged = []
    clusters_total_weight = []
    for points, point_weights in cluster_points:
        total_weight = np.sum(point_weights)
        avg_point = np.average(points, axis=0, weights=point_weights)
        if total_weight < average_filter['min_total_weight']:
            continue
        clusters_averaged.append(avg_point)
        clusters_total_weight.append(total_weight)
    normals = np.array(clusters_averaged)
    weights = np.array(clusters_total_weight)
    normals, _ = normalized(normals)
    idx = np.argsort(weights)[::-1]
    normals = normals[idx]
    weights = weights[idx]

    return normals, weights



from copy import deepcopy
import numpy as np
import open3d as o3d
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from scipy.spatial.transform import Rotation as R
from open3d import __version__ as o3d_version

o3d_major_version = int(o3d_version.split('.')[1])

COLOR_PALETTE = list(map(colors.to_rgb, plt.rcParams['axes.prop_cycle'].by_key()['color']))


def create_line_set(normals_sorted):
    normals_o3d = o3d.utility.Vector3dVector(normals_sorted)
    line_idx = [[i, i + 1] for i in range(normals_sorted.shape[0] - 1)]
    line_idx_o3d = o3d.utility.Vector2iVector(line_idx)
    return o3d.geometry.LineSet(normals_o3d, line_idx_o3d)

def create_open_3d_mesh(triangles, points, triangle_normals=None, color=COLOR_PALETTE[0]):
    """Create an Open3D Mesh given triangles vertices

    Arguments:
        triangles {ndarray} -- Triangles array
        points {ndarray} -- Points array

    Keyword Arguments:
        color {list} -- RGB COlor (default: {[1, 0, 0]})

    Returns:
        mesh -- Open3D Mesh
    """
    mesh_2d = o3d.geometry.TriangleMesh()
    if points.ndim == 1:
        points = points.reshape((int(points.shape[0] / 3), 3))
    if triangles.ndim == 1:
        triangles = triangles.reshape((int(triangles.shape[0] / 3), 3))
        # Open 3D expects triangles to be counter clockwise
        triangles = np.ascontiguousarray(np.flip(triangles, 1))
    mesh_2d.triangles = o3d.utility.Vector3iVector(triangles)
    mesh_2d.vertices = o3d.utility.Vector3dVector(points)
    if triangle_normals is None:
        mesh_2d.compute_vertex_normals()
        mesh_2d.compute_triangle_normals()
    elif triangle_normals.ndim == 1:
        triangle_normals_ = triangle_normals.reshape((int(triangle_normals.shape[0] / 3), 3))
        # triangles = np.ascontiguousarray(np.flip(triangles, 1))
        mesh_2d.triangle_normals = o3d.utility.Vector3dVector(triangle_normals_)
    mesh_2d.paint_uniform_color(color)
    return mesh_2d

def split_triangles(mesh):
    """
    Split the mesh in independent triangles    
    """
    triangles = np.asarray(mesh.triangles).copy()
    vertices = np.asarray(mesh.vertices).copy()

    triangles_3 = np.zeros_like(triangles)
    vertices_3 = np.zeros((len(triangles) * 3, 3), dtype=vertices.dtype)

    for index_triangle, t in enumerate(triangles):
        index_vertex = index_triangle * 3
        vertices_3[index_vertex] = vertices[t[0]]
        vertices_3[index_vertex + 1] = vertices[t[1]]
        vertices_3[index_vertex + 2] = vertices[t[2]]

        triangles_3[index_triangle] = np.arange(index_vertex, index_vertex + 3)

    mesh_return = deepcopy(mesh)
    mesh_return.triangles = o3d.utility.Vector3iVector(triangles_3)
    mesh_return.vertices = o3d.utility.Vector3dVector(vertices_3)
    mesh_return.paint_uniform_color([0.5, 0.5, 0.5])
    return mesh_return


def assign_vertex_colors(mesh, normal_colors, mask=None):
    """Assigns vertex colors by given normal colors
    NOTE: New mesh is returned

    Arguments:
        mesh {o3d:TriangleMesh} -- Mesh
        normal_colors {ndarray} -- Normals Colors

    Returns:
        o3d:TriangleMesh -- New Mesh with painted colors
    """
    split_mesh = split_triangles(mesh)
    vertex_colors = np.asarray(split_mesh.vertex_colors)
    triangles = np.asarray(split_mesh.triangles)
    if mask is not None:
        triangles = triangles[mask, :]
    for i in range(triangles.shape[0]):
        # import ipdb; ipdb.set_trace()
        color = normal_colors[i, :]
        p_idx = triangles[i, :]
        vertex_colors[p_idx] = color

    return split_mesh


def translate_meshes(mesh_list, current_translation=0.0, axis=0):
    translate_amt = None
    translate_meshes = []
    for mesh_ in mesh_list:
        mesh_ = deepcopy(mesh_)

        bbox = mesh_.get_axis_aligned_bounding_box()
        x_extent = bbox.get_extent()[axis]
        translate_amt = [0, 0, 0]
        translate_amt[axis] = current_translation + x_extent / 2.0
        translate_meshes.append(mesh_.translate(translate_amt, relative=False))
        current_translation += x_extent + 0.5
    return translate_meshes

def plot_meshes(*meshes, shift=True, axis_translate=[-2.0, 0, 0]):
    axis = o3d.geometry.TriangleMesh.create_coordinate_frame()
    axis.translate(axis_translate)
    translate_meshes = []
    current_x = 0.0
    if shift:
        for i, mesh in enumerate(meshes):
            inner_meshes = [mesh]
            if isinstance(mesh, list):
                inner_meshes = mesh
            translate_amt = None
            for mesh_ in inner_meshes:
                mesh_ = deepcopy(mesh_)
                if translate_amt is not None:
                    translate_meshes.append(mesh_.translate(translate_amt, relative=True))
                else:
                    bbox = mesh_.get_axis_aligned_bounding_box()
                    x_extent = bbox.get_extent()[0]
                    translate_amt = [current_x + x_extent / 2.0, 0, 0]
                    translate_meshes.append(mesh_.translate(translate_amt, relative=True))
                    current_x += x_extent + 0.5
    else:
        translate_meshes = meshes

    o3d.visualization.draw_geometries([axis, *translate_meshes])

def draw_normals(normals, line_length=0.05):
    normal_tips = normals * (1 + line_length)
    num_lines = normals.shape[0]
    all_points = np.vstack((normals, normal_tips))
    lines = [[i, i + num_lines] for i in range(num_lines)]
    line_set = o3d.geometry.LineSet(
        points=o3d.utility.Vector3dVector(all_points),
        lines=o3d.utility.Vector2iVector(lines),
    )

    return line_set

def calc_angle_delta(mesh, level):
    normals = np.asarray(mesh.triangle_normals)
    v1 = normals[0, :]
    if level == 0:
        v2 = normals[1, :]
    else:
        v2 = normals[3, :]
    diff = v1 @ v2
    deg = np.rad2deg(np.arccos(diff))
    return deg

def normalized(a, axis=-1, order=2):
    """Normalizes a numpy array of points"""
    l2 = np.atleast_1d(np.linalg.norm(a, order, axis))
    l2[l2 == 0] = 1
    return a / np.expand_dims(l2, axis), l2

def normalize_box(a:np.ndarray):
    """Normalizes a 2D numpy array of points into unit box [0,1]"""
    min_x = np.min(a[:, 0])
    max_x = np.max(a[:, 0])
    min_y = np.min(a[:, 1])
    max_y = np.max(a[:, 1])
    range_x = max_x - min_x
    range_y = max_y - min_y
    a[:, 0] = (a[:, 0] - min_x) / range_x
    a[:, 1] = (a[:, 1] - min_y) / range_y

    # print(range_x, range_y, min_x)
    # print(a)
    return a


def vector_magnitude(vec):
    """
    Calculates a vector's magnitude.
    Args:
        - vec (): 
    """
    magnitude = np.sqrt(np.sum(vec**2))
    return(magnitude)


def align_vector_to_another(a=np.array([0, 0, 1]), b=np.array([1, 0, 0])):
    """
    Aligns vector a to vector b with axis angle rotation
    """
    if np.allclose(np.abs(a),np.abs(b),atol=.01):
        return None, None
    axis_ = np.cross(a, b)
    axis_ = axis_ / np.linalg.norm(axis_)
    angle = np.arccos(np.dot(a, b))

    return axis_, angle

def create_arrow(scale=1):
    """
    Create an arrow in for Open3D
    """
    cone_height = scale*0.2
    cylinder_height = scale*0.8
    cone_radius = scale/10
    cylinder_radius = scale/20
    mesh_frame = o3d.geometry.TriangleMesh.create_arrow(cone_radius=cone_radius,
        cone_height=cone_height,
        cylinder_radius=cylinder_radius,
        cylinder_height=cylinder_height)
    return(mesh_frame)

def get_arrow(origin=[0,0,0],end=None,vec=None):
    """
    Creates an arrow from an origin point to an end point,
    or create an arrow from a vector vec starting from origin.
    Args:
        - end (): End point. [x,y,z]
        - vec (): Vector. [i,j,k]
    """
    # print(end)
    scale = 10; beta = 0; gamma = 0
    T = np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]])
    T[:3,-1] = origin
    rotation_3x3 = None
    if end is not None:
        vec = np.array(end) - np.array(origin)
    elif vec is not None:
        vec = np.array(vec)
    if end is not None or vec is not None:
        scale = vector_magnitude(vec)
        mesh = create_arrow(scale)
        axis, angle = align_vector_to_another(b=vec/ scale)
        if axis is None:
            if np.dot(np.array([0, 0, 1]), vec/scale) > 0:
                axis_a = axis
            else:
                rotation_3x3 = np.array([[-1, 0 , 0],
                                        [0, -1, 0],
                                        [0,0, -1]])
        else:
            axis_a = axis * angle
            rotation_3x3 = mesh.get_rotation_matrix_from_axis_angle(axis_a)
    # mesh.transform(T)
    if rotation_3x3 is not None:
        center = [0,0,0] if o3d_major_version > 9 else False
        mesh = mesh.rotate(rotation_3x3, center=center)
    mesh.translate(origin)
    return(mesh)

def get_colors(inp, colormap=plt.cm.viridis, vmin=None, vmax=None):
    norm = plt.Normalize(vmin, vmax)
    return colormap(norm(inp))

def get_pc_all_peaks(peaks, clusters, normals):
    pcd = o3d.geometry.PointCloud()
    peak_normals = np.ascontiguousarray(normals[peaks, :]) * 1.01
    pcd.points = o3d.utility.Vector3dVector(peak_normals)
    colors = get_colors(clusters, colormap=plt.cm.tab20)
    pcd.colors = o3d.utility.Vector3dVector(colors[:, :3])
    return pcd

def get_point_normals(avg_peaks, avg_weights):
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(avg_peaks * 1.01)
    pcd.paint_uniform_color([0, 1, 0])
    meshes = [pcd]

    return meshes

def get_arrow_normals(avg_peaks, avg_weights):
    meshes = []
    for i in range(avg_peaks.shape[0]):
        avg_peak = avg_peaks[i,:]
        mesh = get_arrow(avg_peak, avg_peak * 1.2)
        meshes.append(mesh)

    return meshes


""" Basic demonstation of integrating normals and performing peak detection

"""
import time
import sys
import argparse
import numpy as np
import open3d as o3d

from fastgac import GaussianAccumulatorS2Beta, MatX3d
from fastgac.o3d_util import assign_vertex_colors, plot_meshes, get_colors, create_open_3d_mesh
import matplotlib.pyplot as plt

from tests.python.helpers.setup_helper import cluster_normals, sort_by_distance_from_point
np.random.seed(1)

np.set_printoptions(suppress=True, precision=3)

def integrate_normals_and_visualize(to_integrate_normals, ga):
    to_integrate_normals_mat = MatX3d(to_integrate_normals)
    t0 = time.perf_counter()
    neighbors_idx = np.asarray(ga.integrate(to_integrate_normals_mat))
    t1 = time.perf_counter()
    elapsed_time = (t1 - t0) * 1000
    normalized_counts = np.asarray(ga.get_normalized_bucket_counts())
    color_counts = get_colors(normalized_counts)[:, :3]
    refined_icosahedron_mesh = create_open_3d_mesh(np.asarray(ga.mesh.triangles), np.asarray(ga.mesh.vertices))
    # Colorize normal buckets
    colored_icosahedron = assign_vertex_colors(refined_icosahedron_mesh, color_counts, None)

    return colored_icosahedron

    
def example_normals(normals:np.ndarray):
    kwargs_s2 = dict(level=4)

    axis_frame = o3d.geometry.TriangleMesh.create_coordinate_frame(0.5).translate([-2.0, 0, 0])
    # Create Gaussian Accumulator
    ga_cpp_s2 = GaussianAccumulatorS2Beta(**kwargs_s2)
    # Integrate the normals and get open3d visualization
    colored_icosahedron  = integrate_normals_and_visualize(normals, ga_cpp_s2)
    o3d.visualization.draw_geometries([colored_icosahedron, axis_frame])
    # New simplified API for finding peaks
    res = np.array(ga_cpp_s2.find_peaks(threshold_abs=20, cluster_distance=0.1, min_cluster_weight=0.2))
    print("New Detected Peaks:")
    res = sort_by_distance_from_point(res)
    print(res)


    full_image = np.asarray(ga_cpp_s2.ico_chart.image)
    plt.imshow(full_image)
    plt.xticks(np.arange(0, full_image.shape[1], step=1))
    plt.yticks(np.arange(0, full_image.shape[0], step=1))
    plt.show()

    # Don't forget to reset the GA
    ga_cpp_s2.clear_count()

def main():
    parser = argparse.ArgumentParser(description='Integrate some Normals')
    parser.add_argument('--path', type=str, help='Specify an optional file')
    args = parser.parse_args()
    if args.path is None:
        clusters, normals = cluster_normals(10, 1000, patch_deg=5)
        combined =np.concatenate(clusters)
    else:
        print("loading data from ", args.path)
        data = np.load(args.path)
        combined = data['clusters']
        normals = data['normals']
    print(sort_by_distance_from_point(normals))
    example_normals(combined)

if __name__ == "__main__":
    main()


""" This example demonstrates four (4) different forms of GaussianAccumulators. It also demonstrates two forms of peak detection.
"""
import time
from pathlib import Path
from collections import namedtuple

import numpy as np
import open3d as o3d
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt

from fastgac import GaussianAccumulatorKD, GaussianAccumulatorOpt, GaussianAccumulatorS2Beta, MatX3d, convert_normals_to_hilbert, IcoCharts
from fastgac.peak_and_cluster import find_peaks_from_accumulator, find_peaks_from_ico_charts
from fastgac.helper import down_sample_normals
from fastgac.o3d_util import get_arrow, get_pc_all_peaks, get_arrow_normals, plot_meshes, assign_vertex_colors, create_open_3d_mesh, get_colors, create_line_set
from examples.python.util.mesh_util import get_mesh_data_iterator

from src.Python.slowga import (GaussianAccumulatorKDPy, filter_normals_by_phi)


def get_image_peaks(ga_cpp_s2, level=2, **kwargs):
    ico_chart = IcoCharts(level)
    normalized_bucket_counts_by_vertex = ga_cpp_s2.get_normalized_bucket_counts_by_vertex(True)
    ico_chart.fill_image(normalized_bucket_counts_by_vertex)

    find_peaks_kwargs = dict(threshold_abs=20, min_distance=1, exclude_border=False, indices=False)
    cluster_kwargs = dict(t=0.2, criterion='distance')
    average_filter = dict(min_total_weight=0.05)

    peaks, clusters, avg_peaks, avg_weights = find_peaks_from_ico_charts(ico_chart, np.asarray(
        normalized_bucket_counts_by_vertex), find_peaks_kwargs=find_peaks_kwargs, cluster_kwargs=cluster_kwargs, average_filter=average_filter)
    gaussian_normals_sorted = np.asarray(ico_chart.sphere_mesh.vertices)
    pcd_all_peaks = get_pc_all_peaks(peaks, clusters, gaussian_normals_sorted)
    arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)

    print(avg_peaks)

    return [pcd_all_peaks, *arrow_avg_peaks]


def plot_hilbert_curve(ga: GaussianAccumulatorKDPy, plot=False):

    normals = np.asarray(ga.get_bucket_normals())
    normalized_counts = np.asarray(ga.get_normalized_bucket_counts())
    colors = get_colors(normalized_counts)[:, :3]
    bucket_normals_hv = np.asarray(ga.get_bucket_sfc_values())
    num_buckets = ga.num_buckets
    idx_sort = np.argsort(bucket_normals_hv)
    bucket_normals_hv_sorted = bucket_normals_hv[idx_sort]
    colors = colors[idx_sort, :]
    accumulator_normalized_sorted = normalized_counts[idx_sort]
    gaussian_normals_sorted = normals[idx_sort, :]

    # Find Peaks using 1D signal detector
    peaks, clusters, avg_peaks, avg_weights = find_peaks_from_accumulator(
        gaussian_normals_sorted, accumulator_normalized_sorted)

    # 2D Plots
    if plot:
        class_name_str = type(ga).__name__
        if class_name_str == 'GaussianAccumulatorS2':
            fig, ax = plt.subplots(1, 1, figsize=(5, 5))

        else:
            proj_ = np.asarray(ga.get_bucket_projection())
            proj = proj_[idx_sort, :]
            fig, axs = plt.subplots(2, 1, figsize=(8, 10))
            ax = axs[0]
            scatter1 = ax.scatter(proj[:, 0], proj[:, 1], c=colors, label='Projected Buckets')
            scatter2 = ax.scatter(proj[peaks, :][:, 0], proj[peaks, :][:, 1],
                                  marker='x', c=clusters, label='Clusters', cmap='tab20')
            ax.set_title("Hilbert Curve with Azimuth Equidistant Projection")
            ax.set_xlabel("x*")
            ax.set_ylabel("y*")
            line1 = ax.plot(proj[:, 0], proj[:, 1], c='k', label='Hilbert Curve Connections')[0]
            ax.axis('equal')
            leg = ax.legend(loc='upper left', fancybox=True, shadow=True)

            # we will set up a dict mapping legend line to orig line, and enable
            # picking on the legend line
            lines = [line1, scatter1, scatter2]
            lined = dict()
            for legline, origline in zip(leg.legendHandles, lines):
                legline.set_picker(5)  # 5 pts tolerance
                lined[legline] = origline

            def onpick(event):
                # on the pick event, find the orig line corresponding to the
                # legend proxy line, and toggle the visibility
                legline = event.artist
                origline = lined[legline]
                vis = not origline.get_visible()
                origline.set_visible(vis)
                # Change the alpha on the line in the legend so we can see what lines
                # have been toggled
                if vis:
                    legline.set_alpha(1.0)
                else:
                    legline.set_alpha(0.2)
                fig.canvas.draw()

            ax = axs[1]
            fig.canvas.mpl_connect('pick_event', onpick)

        ax.bar(np.arange(num_buckets), accumulator_normalized_sorted)
        ax.scatter(peaks, accumulator_normalized_sorted[peaks], marker='x', c=clusters, cmap='tab20')

        ax.set_title("Histogram of Normal Counts sorted by Hilbert Values")
        ax.set_xlabel("Hilbert Value (Ascending)")
        ax.set_ylabel("Normal Counts")
        fig.tight_layout()
        plt.show()

    pcd_all_peaks = get_pc_all_peaks(peaks, clusters, gaussian_normals_sorted)
    arrow_avg_peaks = get_arrow_normals(avg_peaks, avg_weights)
    return [pcd_all_peaks, *arrow_avg_peaks]


def visualize_gaussian_integration(ga: GaussianAccumulatorKDPy, mesh: o3d.geometry.TriangleMesh, integrate_kwargs=dict(), **kwargs):
    num_buckets = ga.num_buckets
    to_integrate_normals = down_sample_normals(np.asarray(mesh.triangle_normals))
    num_normals = to_integrate_normals.shape[0]

    class_name_str = type(ga).__name__
    # integrate normals
    if class_name_str in ['GaussianAccumulatorKD', 'GaussianAccumulatorOpt', 'GaussianAccumulatorS2', 'GaussianAccumulatorS2Beta']:
        to_integrate_normals = MatX3d(to_integrate_normals)

    t0 = time.perf_counter()
    neighbors_idx = np.asarray(ga.integrate(to_integrate_normals, **integrate_kwargs))
    t1 = time.perf_counter()
    elapsed_time = (t1 - t0) * 1000
    print("{}; Number of cell in GA: {}; Query Size (K): {}; Execution Time(ms): {:.1f}".format(
        class_name_str, num_buckets, num_normals, elapsed_time))

    # For visualization
    normalized_counts = np.asarray(ga.get_normalized_bucket_counts())
    color_counts = get_colors(normalized_counts)[:, :3]
    refined_icosahedron_mesh = create_open_3d_mesh(np.asarray(ga.mesh.triangles), np.asarray(ga.mesh.vertices))
    # Colorize normal buckets
    colored_icosahedron = assign_vertex_colors(refined_icosahedron_mesh, color_counts, None)

    return colored_icosahedron, np.asarray(to_integrate_normals), neighbors_idx


def main():
    print("Here we are going to try out 4 different types of Gaussian Accumulators")
    print("GaussianAccumulatorKDPy = GA using k-d tree implemented in scipy")
    print("GaussianAccumulatorKD = GA using k-d tree implemented in C++ using nanoflann")
    print("GaussianAccumulatorOpt = GA spacing filling curves and local search. Optimized for top hemisphere. Don't use.")
    print("GaussianAccumulatorS2 = GA spacing filling curves and local search. Works on full sphere. This is the really the best")
    print("")
    kwargs_base = dict(level=4, max_phi=180)
    kwargs_kdd = dict(**kwargs_base, max_leaf_size=10)
    kwargs_opt = dict(**kwargs_base)
    kwargs_s2 = dict(**kwargs_base)

    kwargs_opt_integrate = dict(num_nbr=12)
    # Get an Example Mesh
    ga_py_kdd = GaussianAccumulatorKDPy(**kwargs_kdd)
    ga_cpp_kdd = GaussianAccumulatorKD(**kwargs_kdd)
    ga_cpp_opt = GaussianAccumulatorOpt(**kwargs_opt)
    ga_cpp_s2 = GaussianAccumulatorS2Beta(level=4)

    query_max_phi = kwargs_base['max_phi']

    for i, mesh in enumerate(get_mesh_data_iterator()):
        if i < 0:
            continue

        colored_icosahedron_py, normals, neighbors_py = visualize_gaussian_integration(
            ga_py_kdd, mesh, max_phi=query_max_phi)
        colored_icosahedron_cpp, normals, neighbors_cpp = visualize_gaussian_integration(
            ga_cpp_kdd, mesh, max_phi=query_max_phi)
        colored_icosahedron_opt, normals, neighbors_opt = visualize_gaussian_integration(
            ga_cpp_opt, mesh, max_phi=query_max_phi, integrate_kwargs=kwargs_opt_integrate)
        colored_icosahedron_s2, normals, neighbors_s2 = visualize_gaussian_integration(
            ga_cpp_s2, mesh, max_phi=query_max_phi, integrate_kwargs=kwargs_opt_integrate)

        print("Visualing the mesh and the colorized Gaussian Accumulator of type 'GaussianAccumulatorS2'")
        plot_meshes(colored_icosahedron_s2, mesh)

        # 1D Peak detection
        pcd_cpp_s2 = plot_hilbert_curve(ga_cpp_s2, plot=False)
        # 2D Peak Detection
        pcd_cpp_s2_image = get_image_peaks(ga_cpp_s2, **kwargs_base)

        normals_sorted_proj_hilbert = np.asarray(ga_cpp_opt.get_bucket_normals())
        normals_sorted_cube_hilbert = np.asarray(ga_cpp_s2.get_bucket_normals())

        print("Visualize 1D Peak Detection (Left) and 2D Peak Detection (Right).\n")
        plot_meshes([colored_icosahedron_s2, create_line_set(normals_sorted_cube_hilbert * 1.01), *pcd_cpp_s2],
                    [colored_icosahedron_s2, create_line_set(normals_sorted_cube_hilbert * 1.01), *pcd_cpp_s2_image])

        ga_py_kdd.clear_count()
        ga_cpp_kdd.clear_count()
        ga_cpp_opt.clear_count()
        ga_cpp_s2.clear_count()


if __name__ == "__main__":
    main()



import time
import numpy as np
from scipy.signal import find_peaks
from scipy.cluster.hierarchy import linkage, fcluster
from .helper import normalized


def find_peaks_from_accumulator(gaussian_normals_sorted, accumulator_normalized_sorted,
               find_peaks_kwargs=dict(height=0.05, threshold=None, distance=4, width=None, prominence=0.07),
               cluster_kwargs=dict(t=0.15, criterion='distance')):
    t0 = time.perf_counter()
    peaks, _ = find_peaks(accumulator_normalized_sorted, **find_peaks_kwargs)
    t1 = time.perf_counter()

    gaussian_normal_1d_clusters = gaussian_normals_sorted[peaks,:]
    Z = linkage(gaussian_normal_1d_clusters, 'single')
    clusters = fcluster(Z, **cluster_kwargs)
    t2 = time.perf_counter()

    weights_1d_clusters = accumulator_normalized_sorted[peaks]
    average_peaks, average_weights = average_clusters(gaussian_normal_1d_clusters, weights_1d_clusters, clusters)

    print("Peak Detection - Find Peaks Execution Time (ms): {:.1f}; Hierarchical Clustering Execution Time (ms): {:.1f}".format((t1-t0) * 1000, (t2-t1) * 1000))
    return peaks, clusters, average_peaks, average_weights


def get_point_clusters(points, point_weights, clusters):
    point_clusters = []
    cluster_groups = np.unique(clusters)
    for cluster in cluster_groups:
        temp_mask = clusters == cluster
        point_clusters.append((points[temp_mask, :], point_weights[temp_mask]))
    return point_clusters

def average_clusters(peaks, peak_weights, clusters, average_filter=dict(min_total_weight=0.2)):
    cluster_points = get_point_clusters(peaks, peak_weights, clusters)
    clusters_averaged = []
    clusters_total_weight = []
    for points, point_weights in cluster_points:
        total_weight = np.sum(point_weights)
        avg_point = np.average(points, axis=0, weights=point_weights)
        if total_weight < average_filter['min_total_weight']:
            continue
        clusters_averaged.append(avg_point)
        clusters_total_weight.append(total_weight)
    normals = np.array(clusters_averaged)
    normals, _ = normalized(normals)

    return normals, np.array(clusters_total_weight)




import numpy as np
import scipy.ndimage as ndi
from .rank_order import rank_order

def _get_high_intensity_peaks(image, mask, num_peaks):
    """
    Return the highest intensity peak coordinates.
    """
    # get coordinates of peaks
    coord = np.nonzero(mask)
    intensities = image[coord]
    # Highest peak first
    idx_maxsort = np.argsort(-intensities)
    coord = np.transpose(coord)[idx_maxsort]
    # select num_peaks peaks
    if len(coord) > num_peaks:
        coord = coord[:num_peaks]
    return coord


def _get_peak_mask(image, min_distance, footprint, threshold_abs,
                   threshold_rel):
    """
    Return the mask containing all peak candidates above thresholds.
    """
    if footprint is not None:
        image_max = ndi.maximum_filter(image, footprint=footprint,
                                       mode='constant')
    else:
        size = 2 * min_distance + 1
        image_max = ndi.maximum_filter(image, size=size, mode='constant')
    mask = image == image_max
    if threshold_rel is not None:
        threshold = max(threshold_abs, threshold_rel * image.max())
    else:
        threshold = threshold_abs
    mask &= image > threshold
    return mask


def _exclude_border(mask, exclude_border):
    """
    Remove peaks near the borders
    """
    # zero out the image borders
    for i, excluded in enumerate(exclude_border):
        if excluded == 0:
            continue
        mask[(slice(None),) * i + (slice(None, excluded),)] = False
        mask[(slice(None),) * i + (slice(-excluded, None),)] = False
    return mask


def peak_local_max(image, min_distance=1, threshold_abs=None,
                   threshold_rel=None, exclude_border=True, indices=True,
                   num_peaks=np.inf, footprint=None, labels=None,
                   num_peaks_per_label=np.inf):
    """Find peaks in an image as coordinate list or boolean mask.
    Peaks are the local maxima in a region of `2 * min_distance + 1`
    (i.e. peaks are separated by at least `min_distance`).
    If there are multiple local maxima with identical pixel intensities
    inside the region defined with `min_distance`,
    the coordinates of all such pixels are returned.
    If both `threshold_abs` and `threshold_rel` are provided, the maximum
    of the two is chosen as the minimum intensity threshold of peaks.
    Parameters
    ----------
    image : ndarray
        Input image.
    min_distance : int, optional
        Minimum number of pixels separating peaks in a region of `2 *
        min_distance + 1` (i.e. peaks are separated by at least
        `min_distance`).
        To find the maximum number of peaks, use `min_distance=1`.
    threshold_abs : float, optional
        Minimum intensity of peaks. By default, the absolute threshold is
        the minimum intensity of the image.
    threshold_rel : float, optional
        Minimum intensity of peaks, calculated as `max(image) * threshold_rel`.
    exclude_border : int, tuple of ints, or bool, optional
        If positive integer, `exclude_border` excludes peaks from within
        `exclude_border`-pixels of the border of the image.
        If tuple of non-negative ints, the length of the tuple must match the
        input array's dimensionality.  Each element of the tuple will exclude
        peaks from within `exclude_border`-pixels of the border of the image
        along that dimension.
        If True, takes the `min_distance` parameter as value.
        If zero or False, peaks are identified regardless of their distance
        from the border.
    indices : bool, optional
        If True, the output will be an array representing peak
        coordinates. The coordinates are sorted according to peaks
        values (Larger first). If False, the output will be a boolean
        array shaped as `image.shape` with peaks present at True
        elements.
    num_peaks : int, optional
        Maximum number of peaks. When the number of peaks exceeds `num_peaks`,
        return `num_peaks` peaks based on highest peak intensity.
    footprint : ndarray of bools, optional
        If provided, `footprint == 1` represents the local region within which
        to search for peaks at every point in `image`.  Overrides
        `min_distance`.
    labels : ndarray of ints, optional
        If provided, each unique region `labels == value` represents a unique
        region to search for peaks. Zero is reserved for background.
    num_peaks_per_label : int, optional
        Maximum number of peaks for each label.
    Returns
    -------
    output : ndarray or ndarray of bools
        * If `indices = True`  : (row, column, ...) coordinates of peaks.
        * If `indices = False` : Boolean array shaped like `image`, with peaks
          represented by True values.
    Notes
    -----
    The peak local maximum function returns the coordinates of local peaks
    (maxima) in an image. A maximum filter is used for finding local maxima.
    This operation dilates the original image. After comparison of the dilated
    and original image, this function returns the coordinates or a mask of the
    peaks where the dilated image equals the original image.
    See also
    --------
    skimage.feature.corner_peaks
    Examples
    --------
    >>> img1 = np.zeros((7, 7))
    >>> img1[3, 4] = 1
    >>> img1[3, 2] = 1.5
    >>> img1
    array([[0. , 0. , 0. , 0. , 0. , 0. , 0. ],
           [0. , 0. , 0. , 0. , 0. , 0. , 0. ],
           [0. , 0. , 0. , 0. , 0. , 0. , 0. ],
           [0. , 0. , 1.5, 0. , 1. , 0. , 0. ],
           [0. , 0. , 0. , 0. , 0. , 0. , 0. ],
           [0. , 0. , 0. , 0. , 0. , 0. , 0. ],
           [0. , 0. , 0. , 0. , 0. , 0. , 0. ]])
    >>> peak_local_max(img1, min_distance=1)
    array([[3, 2],
           [3, 4]])
    >>> peak_local_max(img1, min_distance=2)
    array([[3, 2]])
    >>> img2 = np.zeros((20, 20, 20))
    >>> img2[10, 10, 10] = 1
    >>> peak_local_max(img2, exclude_border=0)
    array([[10, 10, 10]])
    """
    out = np.zeros_like(image, dtype=bool)

    threshold_abs = threshold_abs if threshold_abs is not None else image.min()

    if isinstance(exclude_border, bool):
        exclude_border = (min_distance if exclude_border else 0,) * image.ndim
    elif isinstance(exclude_border, int):
        if exclude_border < 0:
            raise ValueError("`exclude_border` cannot be a negative value")
        exclude_border = (exclude_border,) * image.ndim
    elif isinstance(exclude_border, tuple):
        if len(exclude_border) != image.ndim:
            raise ValueError(
                "`exclude_border` should have the same length as the "
                "dimensionality of the image.")
        for exclude in exclude_border:
            if not isinstance(exclude, int):
                raise ValueError(
                    "`exclude_border`, when expressed as a tuple, must only "
                    "contain ints."
                )
            if exclude < 0:
                raise ValueError(
                    "`exclude_border` cannot contain a negative value")
    else:
        raise TypeError(
            "`exclude_border` must be bool, int, or tuple with the same "
            "length as the dimensionality of the image.")

    # no peak for a trivial image
    if np.all(image == image.flat[0]):
        if indices is True:
            return np.empty((0, image.ndim), np.int)
        else:
            return out

    # In the case of labels, call ndi on each label
    if labels is not None:
        label_values = np.unique(labels)
        # Reorder label values to have consecutive integers (no gaps)
        if np.any(np.diff(label_values) != 1):
            mask = labels >= 1
            labels[mask] = 1 + rank_order(labels[mask])[0].astype(labels.dtype)
        labels = labels.astype(np.int32)

        # create a mask for the non-exclude region
        inner_mask = _exclude_border(np.ones_like(labels, dtype=bool),
                                     exclude_border)

        # For each label, extract a smaller image enclosing the object of
        # interest, identify num_peaks_per_label peaks and mark them in
        # variable out.
        for label_idx, obj in enumerate(ndi.find_objects(labels)):
            img_object = image[obj] * (labels[obj] == label_idx + 1)
            mask = _get_peak_mask(img_object, min_distance, footprint,
                                  threshold_abs, threshold_rel)
            if exclude_border:
                # remove peaks fall in the exclude region
                mask &= inner_mask[obj]
            coordinates = _get_high_intensity_peaks(img_object, mask,
                                                    num_peaks_per_label)
            nd_indices = tuple(coordinates.T)
            mask.fill(False)
            mask[nd_indices] = True
            out[obj] += mask

        if not indices and np.isinf(num_peaks):
            return out

        coordinates = _get_high_intensity_peaks(image, out, num_peaks)
        if indices:
            return coordinates
        else:
            out.fill(False)
            nd_indices = tuple(coordinates.T)
            out[nd_indices] = True
            return out

    # Non maximum filter
    mask = _get_peak_mask(image, min_distance, footprint, threshold_abs,
                          threshold_rel)

    mask = _exclude_border(mask, exclude_border)

    # Select highest intensities (num_peaks)
    coordinates = _get_high_intensity_peaks(image, mask, num_peaks)

    if indices is True:
        return coordinates
    else:
        nd_indices = tuple(coordinates.T)
        out[nd_indices] = True
        return out



""" Example showing two different forms of space filling curves
"""
import numpy as np

from fastgac import convert_normals_to_s2id, convert_normals_to_hilbert ,GaussianAccumulatorOpt, MatX3d

def main():
    normals = np.array([
        [-1, 0, 0],
        [0, 0, 1],
        [1, 0, 0],
        [0, 1, 0],
        [0, -1, 0]
    ])
    ga = GaussianAccumulatorOpt(level=4)
    proj, hilbert_values = convert_normals_to_hilbert(MatX3d(normals), ga.projected_bbox)
    s2_values = convert_normals_to_s2id(MatX3d(normals))
    print("Normals: ")
    print(normals)
    print("Transformed to uint32 hilbert values (azimuth equal area projection at northpole):")
    print(hilbert_values)
    print("S2ID. Cubic Projection on sphere. Each face has its own hilbert value computed:")
    print(np.asarray(s2_values))



if __name__ == "__main__":
    main()


""" Example script that visualizes the unwrapping process for peak detection 
    The steps here are much more verbose and just meant highlight the methods.
"""
import time
import functools
from pathlib import Path

import numpy as np
from scipy.spatial.transform import Rotation as R
import matplotlib.pyplot as plt
import open3d as o3d

from fastgac import GaussianAccumulatorS2Beta, MatX3d, refine_icosahedron, refine_icochart, IcoCharts
from fastgac.peak_and_cluster import find_peaks_from_ico_charts
from fastgac.o3d_util import get_colors, create_open_3d_mesh, assign_vertex_colors, plot_meshes, translate_meshes
from examples.python.run_meshes import visualize_gaussian_integration

from examples.python.util.mesh_util import get_mesh_data_iterator

def extract_chart(mesh, chart_idx=0):
    triangles = np.asarray(mesh.triangles)
    vertices = np.asarray(mesh.vertices)
    num_triangles = triangles.shape[0]
    chart_size = int(num_triangles / 5)
    chart_start_idx = chart_idx * chart_size
    chart_end_idx = chart_start_idx + chart_size
    triangles_chart = triangles[chart_start_idx:chart_end_idx, :]

    chart_mesh = create_open_3d_mesh(triangles_chart, vertices)
    chart_mesh.vertex_colors = mesh.vertex_colors

    return chart_mesh, chart_start_idx, chart_end_idx


def decompose(ico):
    triangles = np.asarray(ico.triangles)
    vertices = np.asarray(ico.vertices)
    ico_o3d = create_open_3d_mesh(triangles, vertices)
    return triangles, vertices, ico_o3d


def analyze_mesh(mesh):
    """Demonstrates unwrapping and peak detection of a S2 Histogram"""
    LEVEL = 4
    kwargs_opt_integrate = dict(num_nbr=12)

    # Create Gaussian Accumulator
    ga_cpp_s2 = GaussianAccumulatorS2Beta(level=LEVEL)
    # This function will integrate the normals and return an open3d mesh for visualization.
    colored_icosahedron_s2, _, _ = visualize_gaussian_integration(
        ga_cpp_s2, mesh, integrate_kwargs=kwargs_opt_integrate)
    num_triangles = ga_cpp_s2.num_buckets

    # for verification
    ico_s2_organized_mesh = ga_cpp_s2.copy_ico_mesh(True)
    _, _, ico_o3d_s2_om = decompose(ico_s2_organized_mesh)
    colors_s2 = get_colors(range(num_triangles), colormap=plt.cm.tab20)[:, :3]
    colored_ico_s2_organized_mesh = assign_vertex_colors(ico_o3d_s2_om, colors_s2)

    # Demonstrate the five charts for visualization
    bucket_counts = np.asarray(ga_cpp_s2.get_normalized_bucket_counts(True))
    bucket_colors = get_colors(bucket_counts)[:, :3]
    charts_triangles = []
    for chart_idx in range(5):
        chart_size = int(num_triangles / 5)
        chart_start_idx = chart_idx * chart_size
        chart_end_idx = chart_start_idx + chart_size
        icochart_square = refine_icochart(level=LEVEL, square=True)
        _, _, icochart_square_o3d = decompose(icochart_square)
        colored_icochart_square = assign_vertex_colors(
            icochart_square_o3d, bucket_colors[chart_start_idx:chart_end_idx, :])
        charts_triangles.append(colored_icochart_square)

    # Plot the unwrapped icosahedron
    new_charts = translate_meshes(charts_triangles, current_translation=-4.0, axis=1)
    all_charts = functools.reduce(lambda a, b: a + b, new_charts)
    plot_meshes(colored_ico_s2_organized_mesh, colored_icosahedron_s2, all_charts, mesh)
    avg_peaks = np.array(ga_cpp_s2.find_peaks(threshold_abs=25, cluster_distance=0.1, min_cluster_weight=0.15))
    print(avg_peaks)
    full_image = np.asarray(ga_cpp_s2.ico_chart.image)

    plt.imshow(full_image)
    plt.xticks(np.arange(0, full_image.shape[1], step=1))
    plt.yticks(np.arange(0, full_image.shape[0], step=1))
    plt.show()


def visualize_unwrapping():
    """Demonstrate the unwrapping process by color codes sections"""
    LEVEL = 2
    ico = refine_icosahedron(level=0)
    ico_s2 = GaussianAccumulatorS2Beta(level=LEVEL)
    ico_s2_organized_mesh = ico_s2.copy_ico_mesh(True)
    triangles_ico, vertices, ico_o3d = decompose(ico)
    triangles_s2_om, _, ico_o3d_s2_om = decompose(ico_s2_organized_mesh)
    icochart_slanted = refine_icochart(level=LEVEL, square=False)
    _, _, icochart_slanted_o3d = decompose(icochart_slanted)
    icochart_square = refine_icochart(level=LEVEL, square=True)
    _, _, icochart_square_o3d = decompose(icochart_square)

    colors = get_colors(range(triangles_ico.shape[0]), colormap=plt.cm.tab20)[:, :3]
    colors_s2 = get_colors(range(triangles_s2_om.shape[0]), colormap=plt.cm.tab20)[:, :3]

    colored_ico = assign_vertex_colors(ico_o3d, colors)
    colored_ico_s2 = assign_vertex_colors(ico_o3d_s2_om, colors_s2)
    colored_icochart, start_idx, end_idx = extract_chart(colored_ico_s2, chart_idx=0)
    colored_icochart_slanted = assign_vertex_colors(icochart_slanted_o3d, colors_s2[start_idx:end_idx, :])
    colored_icochart_square = assign_vertex_colors(icochart_square_o3d, colors_s2[start_idx:end_idx, :])

    plot_meshes([colored_ico], [colored_ico_s2], colored_icochart, colored_icochart_slanted, colored_icochart_square)


def main():
    visualize_unwrapping()

    for i, mesh in enumerate(get_mesh_data_iterator()):
        if i < 0:
            continue
        analyze_mesh(mesh)


if __name__ == "__main__":
    main()
